/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <errno.h>
 
#include <wtf/Platform.h> // PLATFORM(...)

#if OS(WINDOWS)
/* FIXME: CLEANUP -- because mingw-gcc has its own headers for e.g. <sys/...>,
    we have to avoid adding 'libshim/include' as a -I path, and hence need to
    spell out fuller #include paths for headers that *are* within libshim.
    There must be a better way to organize this.
   -TS-NSA 2012-08-16 */
#include <libshim/include/pthread.h>
#else
#include <pthread.h>
#endif

#include <wkgfx_Misc.h>

#if PLATFORM(WIIU)
	#include <shim_memory.h>
	#include <thread_ctrl.h>
	#include <JSGlobalData.h>

	#include <PlatformOS.h>
	#include <PlatformDefines.h>
	#include <PlatformError.h>
	#include <PlatformSoftwareKeyboard.h>
	#include <PlatformProfiler.h>
	#include <PlatformStatistics.h>
	using namespace LibPlatform;
#else
#if OS(WINDOWS)
#include <windows.h>
#else
#include <sys/stat.h>
#include <mqueue.h>
#include <pthread.h>
#include <fcntl.h>
#endif
#include <dirent.h>
#include <fs_sim.h>

#include "renderer/WiiURenderer.h"
#include "renderer/WiiUWindowType.h"
extern void ForceRedrawWiiUEmuWindows();
bool g_WaitForWebViewRun = false;
bool g_IsInInspectorMode = false;
#endif

#include <wk_appinterface.h>
#include <WebViewWiiU.h>
#include <wtf/MainThread.h>
#include <GraphicsContextPlatformPrivate.h>

#if ENABLE(INSPECTOR_SERVER)
#include <WebInspectorServer.h>
#include <TcpServerWiiU.h>
#include <SocketStreamHandlePrivate.h>
#endif
#if !PLATFORM(WIIU)
// For setting asset-file paths. -TS-NSA 2012-07-11
#include <RenderThemeWiiU.h>
#include <ScrollbarThemeWiiU.h>
#include <InspectorClientWiiU.h>
#endif

#include "JSDOMWindowBase.h"

#include <PlatformMouseEvent.h>
#include <PlatformKeyboardEvent.h>
#include <WiiUEventManager.h>

#include "JSTestEventSystem.h"
#include "../WiiUExtensions/JSSystemEventDispatcher.h"
//#include "../WiiUExtensions/JSIOFileExtension.h"
//#include "../WiiUExtensions/JSSoftwareKeyboardExtension.h"

#include "CharacterData.h"
#include "ExceptionCode.h"
#include "Node.h"

#if !PLATFORM(WIIU)
// FIXME?: coordinates our vsync 
bool g_VSyncMsgSent[WIIUWINDOW_NUM_TYPES] = {0};
WTF::ThreadCondition g_VSyncMsgCond;
WTF::Mutex g_VSyncMsgMutex;
#endif

// -------------------------------------------------------------------------------------------

#if PLATFORM(WIIU) && ENABLE(JIT)
// temporary reference for explicit startup JIT memory
namespace WTF {
void wiiuJIT_StartupMemory(void);
}
#endif

void (*swkbdEventCallback)(void*);

#if PLATFORM(WIIU)
namespace WebCore {
	bool g_AllowScriptExecution = true;
	void (*g_pSendEnableReloadPacket)() = 0;	//Because HNNNNNGGG manual symbol exporting
};
#endif

// -------------------------------------------------------------------------------------------

class WebKitManager
{
public:
	bool WebKitRunStopped() const					{ return !m_bInitialized || (m_bInitialized && m_bWebViewRunStopped); }

	bool GetEnableRedraw(void) const				{ return m_bRedrawEnabled; }
	void SetEnableRedraw(bool bEnable);
	void AcquireGpuFgResources(bool bAcquire)		{ m_acquireGpuFgResources(bAcquire); };
	void SetIsBackground(bool bBackground)			{ m_bBackground = bBackground; }
	bool GetIsBackground(void) const				{ return m_bBackground; }

	void SetPanicExit(void);

	static bool isInitialized()						{ return s_pInstance && s_pInstance->m_bInitialized; }
	static WebViewManager & webViewManager(void)	{ ASSERT(isInitialized()); return s_pInstance->m_webViewManager; }
	static WebKitManager & instance(void)			{ return *s_pInstance; }

	static void pause(void);
	static bool isPaused(void)						{ return s_pInstance && s_pInstance->m_bRunLoopPaused; }
	static void resume(void);
	#if !PLATFORM(WIIU)
	static void workping(void);
	#endif
	static void signalShutdown(void);
	static void waitForShutdown(void);
	static void quit(void);
	static void SetWebkitRestartViews(bool bRestart, float fadeTimeMs, const char * vConfigXML, const unsigned int vConfigXMLLength);
	static void BlockWebkitRestartViews(bool block, int blockSource);
	static bool WebKitRestartViews(); 
	static float WebKitRestartViewsFadeOutTime(void);
	static void WebKitEmptyRenderQueue();

	static void paintCB(const void *texture, WebViewHandle_t webview, void *user_data);

	static bool isWebkitThread()				{ if (!s_pInstance || !s_pInstance->m_bInitialized) return false; return pthread_equal(pthread_self(), s_pInstance->m_workerThread); }

private:
	WebKitManager();
	~WebKitManager();
	
	#if !OS(WINDOWS)
	bool startup(const WebKitCallbacks_t *pCallbacks, signed prio, signed core);
	#else
	bool startup(const WebKitCallbacks_t *pCallbacks, signed prio, signed core, DWORD recvThreadId);
	#endif

	void rendering(void);
	void renderCustomEvents(void *colorBufferP, long width, long height, bool isGamePad);
	void addCustomRenderingCommand(WebKitCustomRenderData *dataP);
	void removeCustomRenderingCommand(WebKitCustomRenderData *dataP);
	void removeCustomRenderingCommandHelper(WebKitCustomRenderData *dataP);

	void paint(const void *pTexture, WebViewHandle_t webView, void *pUserData);

	void callDL(void *pAddr, unsigned size, unsigned long long timeStampValue, unsigned long long *pTimeStamp, bool bDrawSync);

	static void* workerThread(void *pArg);
	void* _workerThread();
	
	static void callDLCB(void *pAddr, unsigned size, unsigned long long *pTimeStamp, unsigned long long timeStampValue, bool bDrawSync);

	void HandleVSync();

	friend bool WebKitStartup(const WebKitCallbacks_t *pCallbacks, signed prio, signed core);
	friend void WebKitShutdown(void);
	friend void WebKitRemoveCustomRenderingCommand(WebKitCustomRenderData *dataP);
	friend void WebKitAddCustomRenderingCommand(WebKitCustomRenderData *dataP);
	friend void WebKitRendering(void);
	friend void WebKitRenderCustomEvents(void *colorBufferP, long width, long height, bool isGamePad);
	friend void WebKitRenderCustomEventsHold(void);
	friend void WebKitRenderCustomEventsRelease(void);
	friend bool WebKitDone(void);
	friend bool WebKitShuttingDown(void);
	
	static WebKitManager *	s_pInstance;				//!< Singleton instance
	
	bool					m_bInitialized;				//!< True if startup() has been called
	bool					m_bRunLoopPaused;			//!< True if WebViewRun is suspended (worker thread may still be running)
	bool					m_bRunLoopQuit;				//!< True if someone has requested the worker thread to stop
	bool					m_bRedrawEnabled;			//!< True if redraw events are processed, false otherwise
	bool					m_bBackground;				//!< True if webkit is logically in BG mode, FG mode otherwise
	bool					m_bPanicExit;				//!< True during a "panic exit" without real shutdown
	volatile bool			m_bWebViewRunStopped;		//!< True if worker thread has ended
	volatile bool			m_bWebViewRunShuttingDown;	//!< True if worker thread is in the process of shutting down, before m_bWebViewRunStopped is true
	#if OS(WINDOWS)
	DWORD					m_recvThreadId;				//!< ID of the message-receiving thread (we post messages to it)
	#endif
	pthread_mutex_t			m_workerMutex;				//!< Protects the above flags from thread contention
	pthread_cond_t			m_workerEvent;				//!< 'Pause' causes the worker thread to wait on this
	pthread_t				m_workerThread;				//!< Worker thread running the main event loop for webkit

	pthread_cond_t			m_workerShutdownEvent;		//!< Used to help communicate between the main thread and the worker thread to coordinate shutdown
	pthread_mutex_t			m_workerShutdownMutex;		//!< Because pthreads. There's no real need for this.

	// for custom rendering
	WebKitCustomRenderData  *m_customRenderingListP;
	pthread_mutex_t			m_customRenderingMutex;
	
	enum {
		RENDER_MSG_MAXNUM = 64,							//!< Maximum number of render messages to queue up
		WEBKIT_STACK_SIZE =	512 << 10,					//!< Huge stack for Webkit. Some layouts can trigger deep recursions!
	};
	
	//! Entry for queue of rendering jobs to do
	struct RenderQueueEntry
	{
		enum TYPE {
			TYPE_CALL_DL = 0,		//!< Job calls a DL (an may sync to GPU afterwards)
			TYPE_SHOW_TEXTURE,		//!< Job pushes current state of render texture to be displayed
		};
			
		TYPE		m_type;			//!< Type of job
		union
		{
			struct
			{
				void *				m_pAddr;			//!< DL start address
				unsigned			m_size;				//!< Size of DL in bytes
				unsigned long long	m_timeStampVal;		//!< Value to use as user timestamp
				unsigned long long *m_pTimeStamp;		//!< Timestamp value pointer for user timestamp
				bool				m_bDrawSync;		//!< True of CPU should wait for GPU after DL execution before raising the event
			} m_callDL;
			
			struct
			{
				const void *	m_pTexture;			//!< Texture to be displayed
				WebViewHandle_t	m_WebView;			//!< Handle of view the texture is representing
				void *			m_pUserData;		//!< User data registered when view was created
			} m_showTexture;
		};
	};

	#if PLATFORM(WIIU)
	// TGM:THESE NEED TO BE ABSTRACTED! 
	OS::MessageQueue				m_renderQueue;
	OS::Message 					m_renderQueueBuffer[RENDER_MSG_MAXNUM];
	#elif OS(WINDOWS)
	// For Windows, we use the thread's built-in message queue
	#else
	mqd_t							m_renderQueue;
	#endif
	RenderQueueEntry				m_renderQueueEntries[RENDER_MSG_MAXNUM];
	unsigned						m_renderQueueEntryIndex;

	WebViewManager					m_webViewManager;

	WebKitPaintCB_t					(*m_appPaint);				//!< Paint callback implemented by application
	WebKitAcquireGpuFgResourcesCB_t	(*m_acquireGpuFgResources);	//!< Signal use / end of use of GPU resources in foreground mode
};

WebKitManager *	WebKitManager::s_pInstance = NULL;

// -------------------------------------------------------------------------------------------

static TestListener 		test_listener;

static WebKitReallocCB_t	(*s_wkRealloc);			//!< User realloc vector (needed as the OS memory vectors do not include a realloc!)

// -------------------------------------------------------------------------------------------

void DispatchSystemEvent(const char *event_type, void *data) {
	JSEvent js_event(WIIU_EMPTY_EVENT, event_type, data); 
	JSSystemEventDispatcher::GetInstance()->DispatchEvent(js_event);
}

void DispatchTestInputEvents(void) {
	test_listener.ListenToRemote();
	test_listener.ListenToNunChuck();
	test_listener.ListenToClassic();
	test_listener.ListenToGamePad();
}

void DispatchInputEvents(void) {
	JSSystemEventDispatcher::GetInstance()->DispatchInputEvents();
}

void WebKitOutOfMemory(void (*callback)(void)) {
	// Send an event to JS side so they can clean up
	JSSystemEventDispatcher::GetInstance()->SetCrashCallback(callback);
	JSSystemEventDispatcher::GetInstance()->SetCrash(true);
}

// -------------------------------------------------------------------------------------------
/**
	Get the view rendering to the TV
	
	@return	Handle for view rendering to the TV
*/
WebViewHandle_t WebViewTV(void)
{
	return (WebViewHandle_t)WebKitManager::webViewManager().GetTVWebView();
}

// -------------------------------------------------------------------------------------------
/**
	Find out if the specified view renders to one of the GamePads
	
	@param	index	Index of GamePad the view should render to
	
	@return Handle for view rendering to GamePad with given index
*/
WebViewHandle_t WebViewGamePad(unsigned int index)
{
	return (WebViewHandle_t)WebKitManager::webViewManager().GetGamePadWebView(index);
}

// -------------------------------------------------------------------------------------------
/**
	Get the view rendering to the inspector
	
	@return Handle for view rendering to the inspector
*/
WebViewHandle_t WebViewInspector(void)
{
	return (WebViewHandle_t)WebKitManager::webViewManager().GetInspectorWebView();
}

// -------------------------------------------------------------------------------------------
/**
	Get an intentionally-ambiguous view, for cases where the precise handle doesn't matter
	
	@return Handle for view rendering to the TV
*/
WebViewHandle_t WebViewAny(void)
{
	return WebViewTV();
}

// -------------------------------------------------------------------------------------------
/**
	Determine the type (TV vs. GamePad) of a given web view
	
	@param	webview	Handle to the view in question
	
	@return Type of the given view
*/
WebViewType_t WebViewType(WebViewHandle_t webview)
{
	switch(WebKitManager::webViewManager().GetDisplayType((WebView *)webview))
	{
		case WebView::TV_DISPLAY_TYPE:
			return WEBVIEW_TV;
		case WebView::GAMEPAD_DISPLAY_TYPE:
			return WEBVIEW_GAMEPAD;
		default:
			return WEBVIEW_NONE;
	}
}

// -------------------------------------------------------------------------------------------
/**
	Get the total count of instantiated views (TV + some number of GamePads)
	
	@return Total number of current views
*/
int NumWebViews(void)
{
	return WebKitManager::webViewManager().GetNumWebViews();
}

// -------------------------------------------------------------------------------------------
/**
	Get the total count of instantiated views (TV + some number of GamePads)
	
	@return Total number of current GamePad views
*/
int NumGamePadWebViews(void)
{
	return WebKitManager::webViewManager().GetNumGamePadWebViews();
}

// -------------------------------------------------------------------------------------------
/**
	Find out if the specified view renders to the TV
*/
bool WebViewRendersToTV(WebViewHandle_t view)
{
	WebView::DisplayType display_type = WebKitManager::webViewManager().GetDisplayType((WebView *)view);
	return display_type == WebView::TV_DISPLAY_TYPE;
}

int WebViewRendersToLocation(WebViewHandle_t view)
{
    WebView::DisplayType display_type = WebKitManager::webViewManager().GetDisplayType((WebView *)view);
    return (int)display_type;
}

// -------------------------------------------------------------------------------------------
/**
	Get the page associated with the given view
*/
void *WebViewPage(WebViewHandle_t view)
{
	WebView *web_view = (WebView *)view;
	return (void *)web_view->getPage();
}

// -------------------------------------------------------------------------------------------
/**
	Make the specified view visible
*/
void WebViewSetVisible(WebViewHandle_t view, bool visible)
{
	WebView *web_view = (WebView *)view;
	web_view->setVisible(visible);
}

// -------------------------------------------------------------------------------------------
/**
	Is this webview currently visible?
*/
bool WebViewIsVisible(WebViewHandle_t view)
{
	WebView *web_view = (WebView *)view;
	return web_view->isVisible();
}

// -------------------------------------------------------------------------------------------
/**
	Return width of the given view
*/
unsigned WebViewGetWidth(WebViewHandle_t view) 
{
	WebView *web_view = (WebView *)view;
	return web_view->width();
}

// -------------------------------------------------------------------------------------------
/**
	Return height of the given view
*/
unsigned WebViewGetHeight(WebViewHandle_t view) 
{
	WebView *web_view = (WebView *)view;
	return web_view->height();
}

// -------------------------------------------------------------------------------------------
/**
	Save a screen shot of the specified view to the specified file
*/
void WebViewSaveScreen(WebViewHandle_t view, const char *filename, bool extendedSaveGamePad)
{
	WebView *webView = (WebView *) view;
	webView->saveScreenshot(filename, extendedSaveGamePad);
}

// -------------------------------------------------------------------------------------------
/**
	Get a copy of the current webview's image buffer
*/
void *WebViewCaptureScreen(WebViewHandle_t view, int dstWidth, int dstHeight, int srcX, int srcY, int srcWidth, int srcHeight)
{
	WebView *webView = (WebView *) view;
	return (void*) webView->captureScreen(dstWidth, dstHeight, srcX, srcY, srcWidth, srcHeight);
}


/**
	Get the URL that a WebView is currently displaying. 
	@param A \a WebViewHandle_t to get the URL from. 
	@return A \a CString that contains the URL
*/
CString WebViewGetURL(WebViewHandle_t view) 
{
	WebView *webView = (WebView *) view;
	return webView->getURLCharacters();
}

// -------------------------------------------------------------------------------------------
/**
*/
bool WebViewsInitialized(void) {
	return WebKitManager::isInitialized();
}

// -------------------------------------------------------------------------------------------
/**
	Service function to create a view for TV and one or more views for GamePad use
*/
void WebViewsCreate(WebViewCallbacks_t *callbacks, WebViewSettings_t *settings, unsigned int gamepad_web_view_count, WebViewInspectorLocation_t  inspectorLocation)
{
	WebKitManager::webViewManager().CreateWebViews(callbacks, settings, gamepad_web_view_count, inspectorLocation);
}

/**
	Returns the callback structured used to create the initial WebViews
	@return A pointer to the \a WebViewCallbacks_t used to initialize the WebViews
*/
WebViewCallbacks_t *WebViewGetCallbacks(void)
{
	return WebKitManager::webViewManager().getCallbacks();
}

/**
	Returns the settings used to create the initial WebViews
	@return A pointer to the \a WebViewSettings_t structure used to initialize the WebViews
*/
WebViewSettings_t *WebViewGetSettings(void)
{
	return WebKitManager::webViewManager().getSettings();
}

/**
	Returns the location value for the WebInspector WebView
	@return A \a WebViewInspectorLocation_t value for the WebInspector WebView
*/
WebViewInspectorLocation_t WebViewGetInspectorLocation(void)
{
	return WebKitManager::webViewManager().getLocation();
}

/**
	Sets the restart web views flag signalling the tear down and recreation of the web views
	@param A \a bool value that is true if the webviews should be restarted and false otherwise
*/
void WebViewSetRestartWebViews(bool restart, float fadeOutTimeMS, const char * vXMLConfigData, const unsigned int vXMLConfigLength)
{
	WebKitManager::webViewManager().setRestartWebViews(restart, fadeOutTimeMS, vXMLConfigData, vXMLConfigLength); 
}

namespace WebCore{
	#if ENABLE(INSPECTOR_SERVER)
	void AddWebInspectorServerReference();
	void RemoveWebInspectorServerReference();
	#endif
	void (*g_pAddWebInspectorServerReference)() = 0;
	void (*g_pRemoveWebInspectorServerReference)() = 0;
}

/**
	Returns the remote inspector frontend URL
	@return A \a CString that is the inspector frontend URL
*/
CString WebInspectorFrontendUrl(void)
{
#if ENABLE(INSPECTOR_SERVER)
	if (WebCore::WebInspectorServer::HasInstance() && !WebCore::WebInspectorServer::shared().getFrontendUrl().isEmpty())
		return WebCore::WebInspectorServer::shared().getFrontendUrl().utf8();
#endif
	return CString("");
}

// -------------------------------------------------------------------------------------------
/**
*/
void EmptyWebKitEventQueue()
{
	//
	// Event queue might have old webview events in it with dangling pointers to deleted webviews...
	//

	WebCore::WiiUEventManager *event_manager = WebCore::WiiUEventManager::instance();
	OS::Report("WebCore::WiiUEventManager::instance()\n");

	// Flush events down the terlet!
	#ifndef NDEBUG
	OS::Report("EmptyWebKitEventQueue: Peeking!\n");
	#endif

	WiiUBaseEventData *data;
	while(event_manager->getNextEvent(&data))
	{
		#ifndef NDEBUG
		OS::Report("EmptyWebKitEventQueue: Event type: %d\n", data->type());
		#endif

		// If a timer has been fired then we need to execute it's shared timer function.
		// If we dont do this nothing will load upon restart, because there will be lingering timers.
		if(data->type() == WIIU_TIMER_FIRED)
		{
			WiiUTimerFired *pTimerFired = static_cast<WiiUTimerFired*>(data);
			static void (*sharedTimerFiredFunction)();
			sharedTimerFiredFunction = (void (*)()) pTimerFired->pFunction;
			if(sharedTimerFiredFunction)
				sharedTimerFiredFunction();
		}

		if (data->type() == WIIU_SET_CERTIFICATE)
		{
			WiiUSetCertificateEvent *cert_data = static_cast<WiiUSetCertificateEvent*>(data);
			if (cert_data->bClientCert)
				WebView::setClientCertificate(cert_data->cert.c_str());
			else
				WebView::setCertificate(cert_data->cert.c_str());
		}

		delete data;
	}
}

// -------------------------------------------------------------------------------------------
/**
	Service function to destory commonly used views
*/
void WebViewsDestroy(void)
{
	OS::Report("WebViewsDestroy Starting\n");

	// Destory views...
	WebKitManager::webViewManager().DestroyWebViews(false);
	OS::Report("webViewManager().DestroyWebViews()\n");

	//
	// Event queue might have old webview events in it with dangling pointers to deleted webviews...
	//

	EmptyWebKitEventQueue();
	WebKitManager::WebKitEmptyRenderQueue();
	OS::Report("WebViewsDestroy Finished.\n");
}

// -------------------------------------------------------------------------------------------
/**
	Send message to webkit thread to create the specified view
	
	@note	This function only returns once the view has been created
*/
WebViewHandle_t WebViewCreate(WebViewSize_t *size, void *surface, WebViewCallbacks_t *cbs, WebViewSettings_t *settings, bool noSetup, void *userdata, WebViewHandle_t inspectorview, bool inactiveDueToExtDisplay)
{
	//return new WebView(size->width, size->height, cbs, WebKitManager::paintCB, settings, surface, noSetup, userdata, (WebView*)inspectorview);

	WebView *webview;
	pthread_cond_t cond;
	pthread_mutex_t mutex;
	pthread_cond_init(&cond, NULL);
	pthread_mutex_init(&mutex, NULL);

	OS::Report("[WebViewCreate]Sending WiiUCreateViewEvent\n");
	WiiUCreateViewEvent* data = new WiiUCreateViewEvent(size, cbs, WebKitManager::paintCB, settings, noSetup, userdata, &webview, &cond, &mutex, (WebView *) inspectorview, inactiveDueToExtDisplay);

   	WebKitSendEvent(data);

	OS::Report("[WebViewCreate]Waiting for WiiUCreateViewEvent\n");
	pthread_mutex_lock(&mutex);
	pthread_cond_wait(&cond, &mutex);
	pthread_mutex_unlock(&mutex);

	pthread_cond_destroy(&cond);
	pthread_mutex_destroy(&mutex);

	OS::Report("[WebViewCreate]WebView created\n");
    return (WebViewHandle_t)webview;
}

// -------------------------------------------------------------------------------------------
/**
	Send message to webkit thread to delete the given view

	@param	view				View to be destroyed
	@param	bOnWebkitThread		False for general use, true only if call is issued from main webkit handling thread
	
	@note	This function only returns once the view has been deleted
*/
void WebViewDestroy(WebViewHandle_t view, bool bOnWebkitThread)
{
	// If we are on the webkit thread, we need to do this "immediately"...
	if (bOnWebkitThread)
	{
		OS::Report("[WebViewDestroy]Beginning Cache Clear for view 0x%X\n", (unsigned int)view);
		// Yes. Let's do it!
		((WebView*)view)->clearCaches();
		//((WebView*)view)->clearAllCookies();	//NSA-RW 01.23.2013 -- This is removing persistent cookies from disk, which is not a thing we want to happen.

		OS::Report("[WebViewDestroy]Clear Complete. Deleting view.\n");
		delete (WebView*) view;
		OS::Report("[WebViewDestroy]View Deleted.\n");
		return;
	}

	// From another thread. Better send a message...

	// Setup conditional variable to allow us to wait for the view to be destroyed by the webkit thread
	pthread_cond_t cond;
	pthread_mutex_t mutex;
	pthread_cond_init(&cond, NULL);
	pthread_mutex_init(&mutex, NULL);
	
	OS::Report("[WebViewDestroy]Asking someone else to clean up webview 0x%X\n", (unsigned int)view);
	// Send out the message to destroy the view
	WiiUDestroyViewEvent* data = new WiiUDestroyViewEvent(view, &cond, &mutex);
   	WebKitSendEvent(data);
	
	OS::Report("[WebViewDestroy]Waiting for cleanup to occur.\n");

	// Wait for it...
	pthread_mutex_lock(&mutex);
	pthread_cond_wait(&cond, &mutex);
	pthread_mutex_unlock(&mutex);

	// Cleanup!
	pthread_cond_destroy(&cond);
	pthread_mutex_destroy(&mutex);

	OS::Report("[WebViewDestroy]WebView Cleanup complete.\n");
}

// -------------------------------------------------------------------------------------------
/**
	Trigger the runloop of webkit to temporarily stop
*/
void WebKitPause(void)
{
	WebKitManager::pause();
}

// -------------------------------------------------------------------------------------------
/**
	Resume from WebViewPause
*/
void WebKitResume(void)
{
	WebKitManager::resume();
}

#if !PLATFORM(WIIU)
// -------------------------------------------------------------------------------------------
/**
	Check if webkit is paused
*/
bool WebKitIsPaused(void)
{
	return WebKitManager::isPaused();
}

// -------------------------------------------------------------------------------------------
/**
	Wake up/signal the work thread, without actually pausing or resuming anything
*/
void WebKitWorkPing(void)
{
	WebKitManager::workping();
}
#endif

// -------------------------------------------------------------------------------------------
/**
	Trigger the runloop of webkit to exit out
*/
void WebKitQuit(void)
{
	WebKitManager::quit();
}

// -------------------------------------------------------------------------------------------
/**
	Set user agent string with view
*/
void WebViewSetUA(WebViewHandle_t view, const char *agent)
{
	WiiUSetUserAgentEvent* data = new WiiUSetUserAgentEvent(view, std::string(agent));
   	WebKitSendEvent(data);
}

// -------------------------------------------------------------------------------------------
/**
	Set custom header string with view
*/
void WebViewSetCustomHeader(WebViewHandle_t view, const char *header)
{
	WiiUSetCustomHeaderEvent* data = new WiiUSetCustomHeaderEvent(view, std::string(header));
   	WebKitSendEvent(data);
}

// ----------------------------------------------------------------------------------------------
/**
	Set proxy information for webkit
*/
void WebKitSetProxyInfo(const char *host, unsigned long port, WebKitProxyType type, const char *username, const char *password)
{
	WiiUSetProxyInfoEvent* data = new WiiUSetProxyInfoEvent(username, password, host, port, type);
	WebKitSendEvent(data);
}

// ----------------------------------------------------------------------------------------------
/**
	Set SSL CA certificate for webkit
*/
void WebKitSetCertificate(const char *cert)
{
	WiiUSetCertificateEvent *data = new WiiUSetCertificateEvent(cert);
	WebKitSendEvent(data);
}

// ----------------------------------------------------------------------------------------------
/**
	Set SSL client certificate for webkit
*/
void WebKitSetClientCertificate(const char *cert)
{
	WiiUSetCertificateEvent *data = new WiiUSetCertificateEvent(cert);
	data->bClientCert = true;
	WebKitSendEvent(data);
}

// -------------------------------------------------------------------------------------------
/**
	Get the TCP port bound to the remote inspector server
*/
int WebInspectorPort(void)
{
   	#if ENABLE(INSPECTOR_SERVER)
	if (WebCore::WebInspectorServer::HasInstance())
   		return WebCore::WebInspectorServer::shared().getPort();
	return 0;
   	#else
   	return 0;
   	#endif
}

// -------------------------------------------------------------------------------------------
/**
	Load new URL to view
*/
void WebViewLoad(WebViewHandle_t view, const char *url)
{
	WiiULoadURLEvent *data = new WiiULoadURLEvent(view, std::string(url));
   	WebKitSendEvent(data);
}

// ----------------------------------------------------------------------------------------------
/**
	Send messages to the webkit thread
*/
int WebKitSendEvent(WiiUBaseEventData *pData)
{
	WiiUEventType_t type = pData->type();

	//If we're paused or the software keyboard is open in any capacity when we want to restart, we'll
	//need to turn event processing back on and send the restart event
	if (type == WIIU_RESTART_WEBVIEWS && (WebKitManager::isPaused() || LibPlatform::SoftwareKeyboard::GetInstance()->IsDisplaying()))
	{
		LibPlatform::SoftwareKeyboard * SWKBD = LibPlatform::SoftwareKeyboard::GetInstance();

		//If the software keyboard is open, close it
		if (SWKBD->IsDisplaying())
			SWKBD->Dismiss(LibPlatform::SoftwareKeyboard::DISMISS_CANCEL);

		//Start the event loop back up
		if (WebKitManager::isPaused())
			WebKitResume();

		//Send the restart event
		WebView::sendEvent(pData);
		return 0;
	}

	if (WebKitManager::isInitialized() && (!WebKitManager::isPaused() || type == WIIU_SHUTDOWN_NOTIFICATION || type == WIIU_BG_NOTIFICATION || type == WIIU_FG_NOTIFICATION))
		WebView::sendEvent(pData);
	else
		delete pData;
	return 0;
}

// ----------------------------------------------------------------------------------------------
/**
*/
WEBKIT_API int WebKitSendEventVSync(WebViewHandle_t view)
{
	return WebKitSendEvent(new WiiUDisplayVsyncEvent(view));
}
// ----------------------------------------------------------------------------------------------
/**
*/
WEBKIT_API int WebKitSendEventInputUpdate()
{
    return WebKitSendEvent(new WiiUInputUpdate());
}

WEBKIT_API int WebKitSendEventMouse(int eventType, WebViewHandle_t view, int x, int y, int button, int state, int clickCount)
{
   return WebKitSendEvent(new WiiUMouseData(eventType, view, x, y, button, state, clickCount));
}

WEBKIT_API void WebKitSendEventUSBKeyboard(bool attached) {
	WebKitSendEvent(new WiiUUSBKeyboardData(attached));
}

// ----------------------------------------------------------------------------------------------

void WebViewEnableInspector(WebViewHandle_t view, bool enable)
{
    WebView *webView = (WebView *) view;
    WiiUWebInspectorEnableEvent *data = new WiiUWebInspectorEnableEvent(webView, enable);
    WebKitSendEvent(data);
}

// ----------------------------------------------------------------------------------------------

void WebViewEnableRemoteInspector(int port, std::string & frontendUrl, std::string & hostUrl)
{
    #if ENABLE(INSPECTOR_SERVER)
    WiiUWebInspectorServerStartEvent * data = new WiiUWebInspectorServerStartEvent(port, frontendUrl, hostUrl);
    WebKitSendEvent(data);
    #endif
}
	
#if PLATFORM(WIIU)
// ----------------------------------------------------------------------------------------------
/**
	Poll for and initiate reactions of this thread to a process switch message
*/
static void WebViewHandleProcessSwitch()
{
	switch(shim_PollForSwitch())
	{
		case	SHIM_SWITCH_TO_BG:
		{
			//
			// Going to BG mode!
			//

			// Send notification to javascript that we are going to the background
			WebKitSendEvent(new WiiUBackgroundNotification());

			// Now reenter the event loop for a bit, once the above is done we'll continue the process in 
			// WebViewProcessSwitchNotifyDone()

			break;
		}
		case	SHIM_SWITCH_TO_FG:
		{
			//
			// Going to FG mode!
			//

			// We need the FG resources back!
			WebKitManager::instance().AcquireGpuFgResources(true);

			shim_SetHaveThreadsBeenMoved(false);

			// Now we can allow redraw events again
			WebKitManager::instance().SetEnableRedraw(true);

			// Get JIT going again
#if ENABLE(JIT)
			JSC::JSGlobalData::setLockJIT(false);
#endif

			// Make sure we're ready for gaining the GPU, write gather etc.
			GraphicsContextPlatformPrivate::prepareForProcessSwitch(true);

			// Flag to the "greater public" (e.g. the main application thread) that we're logically in FG mode now!
			WebKitManager::instance().SetIsBackground(false);

			// acknowledge will be send via function below once WK notification is done

			// Send notification that we are going to the foreground
			WebKitSendEvent(new WiiUForegroundNotification());

			// Now reenter the event loop for a bit, once the above is done we'll continue the process in 
			// WebViewProcessSwitchNotifyDone()
			break;
		}
		case	SHIM_SWITCH_SHUTDOWN:
		{
			//
			// Going to shutdown!
			//

			// Disable rendering until shutdown is complete. We must not touch the GPU!
			GraphicsContextPlatformPrivate::setDrawEnabled(false);

			// Get JIT going again
			//
			// Notes:
			//
			// - We are in BG mode, but on the right core non the less! Special case for shutdown!
			// - We must get JIT going in case webkit blocked on entering JS code as we switched to BG mode before this! Otherwise we would never exit!
			// - Above switch to disable issuing of any graphics calls will safely circumvent the issue that JS code could still directly or indirectly trigger rendering in BG mode!
			//
#if ENABLE(JIT)
			JSC::JSGlobalData::setLockJIT(false);
#endif

			// Send notification that we're about to shutdown
			WebKitSendEvent(new WiiUShutdownNotification());

			// Now reenter the event loop for a bit, once the above is done we'll continue the process in 
			// WebViewProcessSwitchNotifyDone()

			break;
		}
	}
}

// ----------------------------------------------------------------------------------------------
/**
	Finish of reaction of this thread to process switch messages
*/
static void WebViewProcessSwitchNotifyDone(unsigned int shimSwitchMessage)
{
	switch(shimSwitchMessage)
	{
		case	SHIM_SWITCH_TO_BG:
		{
			//
			// Going to BG mode!
			//

			// Switch JIT off
			// (this must wait for any WebWorkers to finish execution -> can take a while)
#if ENABLE(JIT)
			JSC::JSGlobalData::setLockJIT(true);
#endif

			// Make sure we don't process painjt messages even if we get them
			WebKitManager::instance().SetEnableRedraw(false);

			// Make sure we're ready for loosing GPU, write gather etc.
			GraphicsContextPlatformPrivate::prepareForProcessSwitch(false);

			// Any FG resource usage is done from here
			WebKitManager::instance().AcquireGpuFgResources(false);

			// Flag to the "greater public" (e.g. the main application thread) that we're logically in BG mode now!
			WebKitManager::instance().SetIsBackground(true);
            shim_SetHaveThreadsBeenMoved(true);

			// Tell shim that we're ready to go!
			shim_AcknowledgeSwitch();
			break;
		}
		case	SHIM_SWITCH_TO_FG:
		{
			//
			// Going to FG mode!
			//

			// Tell shim we're ready!
			shim_AcknowledgeSwitch();
			break;
		}
		case	SHIM_SWITCH_SHUTDOWN:
		{
			//
			// Going to shutdown!
			//

			// Tell shim we're ready!
			shim_AcknowledgeSwitch();
			break;
		}
	}
}

#endif // PLATFORM(WIIU)

// ----------------------------------------------------------------------------------------------
/**
	Web View main loop
*/
// Fixme: While this loop doesn't actually make sense because the WebViewRun() function is already called inside of a loop,
//         the two events (WIIU_INTERRUPT_WEBVIEW and WIIU_STOP_EVENT_LOOP) that can actually return false from WebKitProcessEvents()
//         are effectively ignored because the _workerThread() loop will just call WebViewRun() again.
//        Hacking around this by adding this global for now, but the event loop should really be looked at again later. -Nsa-LH 6/8/12
static bool g_actuallyQuitEventLoop = false;
static int WebViewRun()
{
	//unsigned int DEBUG_Counter = 0;
	//unsigned int DEBUG_ShutdownTime = 0;

	while (WebKitProcessEvents())
	{
		LibPlatform::Profiler::HeartBeatMain0();
		//DEBUG -- Wait some time, then send a shutdown event.
		/*////////////////////////////////////////////////////////////////////////
		DEBUG_Counter++;
		if (!(DEBUG_Counter % 1000))
		{
			DEBUG_Counter++;
			OS::Report("Shutting down in T-%u\n", 5 - DEBUG_ShutdownTime);
		}

		if (DEBUG_ShutdownTime >= 5)
		{
			DEBUG_ShutdownTime = 0;
			WebKitSendEvent(new WiiUStopEventLoopEvent());
		}
		////////////////////////////////////////////////////////////////////////*/
	}

	return g_actuallyQuitEventLoop;
}

// ----------------------------------------------------------------------------------------------
/**
*/
bool WebKitProcessEvents(void)
{
    //TODO: Verify this is the main thread
	static bool DisableRedraw = false;
	WiiUBaseEventData *	data;
	WebView *			view = NULL;

	WebCore::WiiUEventManager *cemInst = WebCore::WiiUEventManager::instance();

	#if PLATFORM(WIIU)
	do
		WebViewHandleProcessSwitch();
	while(!cemInst->waitEventWithTimeout(&data, 0.020));	// 20ms timeout
	#else
	cemInst->waitEvent(&data);
	#endif

	if (data == NULL)
		return true;
	
	WiiUEventType_t type = data->type();

    if(type == WIIU_INTERRUPT_WEBVIEW)
    {
        delete data;
        return false;
    }

	//If we're asked to stop running the event loop altogether (causing the
	//thread running the function to exit), go ahead and do that.
	if (type == WIIU_STOP_EVENT_LOOP)
	{
	    delete data;
		g_actuallyQuitEventLoop = true;
		return false;
	}

	//If we're trying to exit, we don't care about any events other than
	//the one to destroy the WebViews.
	#if !PLATFORM(WIIU)
	if (g_WaitForWebViewRun && (type != WIIU_DESTROY_VIEW))
		return true;
	#endif

	switch (type)
	{
		case WIIU_PREPARE_FOR_RESTART:
		{
			WiiUPrepareForRestartEvent * PrepEvent = (WiiUPrepareForRestartEvent*)data;

			//Tell the other thread we're processing this event
			printf("[PT] Prepared for restart, signaling main thread\n");
			pthread_cond_signal(PrepEvent->m_pSignal);
			*(PrepEvent->m_pProcessed) = true;

			DisableRedraw = true;
			WebCore::g_AllowScriptExecution = false;
		} break;

		case WIIU_RESTART_FINISHED:
		{
			WiiURestartFinishedEvent * RestartEvent = (WiiURestartFinishedEvent*)data;

			printf("[PT] RestartFinished signal received, signaling main thread.\n");
			pthread_cond_signal(RestartEvent->m_pSignal);

			DisableRedraw = false;
			WebCore::g_AllowScriptExecution = true;

			//Empty the render queue so we don't have extra render events
			//that contain now-invalid data.
			WebKitManager::WebKitEmptyRenderQueue();
		} break;

        case WIIU_TIMER_FIRED:
        {
            WiiUTimerFired *pFiredTimer = static_cast<WiiUTimerFired*>(data);
            static void (*sharedTimerFiredFunction)();
            sharedTimerFiredFunction = (void (*)()) pFiredTimer->pFunction;
            if (sharedTimerFiredFunction) {
                sharedTimerFiredFunction();
            }
            break;
        }
        case WIIU_INPUT_UPDATE:
        {
            JSSystemEventDispatcher::GetInstance()->DispatchInputEvents();
            break;
	    }
        case WIIU_REDRAW:
		{
			//We're not allowed to redraw if we're restarting the webviews
			if (DisableRedraw)
				break;

            WiiURedrawEvent *event_data = static_cast<WiiURedrawEvent*>(data);

            view = (WebView *) event_data->webview;
            if (view != NULL)
            {
                #if !PLATFORM(WIIU)
                static bool LocatedInspector = false;   //TODO -- If WebKit is ever going to be properly restarted, this hack will need to be removed.
                if (view->temp_windowType == GAMEPAD_UNDEFINED_ID)
                {
                    //If we're in inspector mode and we've already located the inspector, ignore redraw events for other windows.
                    if (g_IsInInspectorMode && LocatedInspector)
                    {
                        printf("RENDERING: Skipping redraw event for undefined non-inspector window\n");
                        break;
                    }

                    //printf("--Re-queuing Redraw Event to give the WebView objects a chance to be initialized\n");
                    cemInst->putEvent(data);
                    return true;
                }

                if (!view->m_pNewRenderer)
                {
                    printf("RENDERING: Skipping Redraw event for uninitialized window type 0x%X\n", view->temp_windowType);
                    break;
                }

                if (view->temp_windowType == WIIUWINDOW_INSPECTOR)
                    LocatedInspector = true;
                //printf("RENDERING: Redrawing window type 0x%X\n", view->temp_windowType);
                #endif

                PlatformStatistics::GetInstance()->StopTimer("RenderContext", "RenderTime");
                view->redraw();
                PlatformStatistics::GetInstance()->StartTimer("RenderContext", "RenderTime");
            }
            break;
        }
        case WIIU_DISPLAY_VSYNC:
        {
			//We're not allowed to redraw if we're restarting the webviews
			if (DisableRedraw)
				break;

            WiiUDisplayVsyncEvent *event_data = static_cast<WiiUDisplayVsyncEvent *>(data);

            view = (WebView *) event_data->webview;
            if (view != NULL) {

                #if PLATFORM(WIIU)
                WebView *view = WebKitManager::webViewManager().GetTVWebView();
                if(view != NULL)
                {
                    view->runScheduledAnimations();
                }
                for(int i = 0; i < WebKitManager::webViewManager().GetNumGamePadWebViews(); i++)
                {
                    view = WebKitManager::webViewManager().GetGamePadWebView(i);
                    if (view)
                    {
                        view->runScheduledAnimations();
                    }
                }
                #else
                {
                    if (!event_data->m_RefreshOnly)
                        view->runScheduledAnimations();

                    //We're going to let redraw do its thing in case there's some change
                    //that hasn't happened yet for some reason, then we'll force the paint call
                    view->redraw();
                    view->_ForcePaintCall();

                    g_VSyncMsgMutex.lock();
                    if (view->temp_windowType != (WiiUWindowType)GAMEPAD_UNDEFINED_ID)
                        g_VSyncMsgSent[view->temp_windowType] = 0;  //We've handled the vsync message, we can take another now
                    g_VSyncMsgMutex.unlock();
                    g_VSyncMsgCond.signal();
                }
                #endif
            }
            break;
        }
        case WIIU_MOUSE_MOVE:
        {
            WiiUMouseData *mev_data = static_cast<WiiUMouseData *>(data);

            view = (WebView *) mev_data->webview;
            if (view != NULL) {
                WebCore::PlatformMouseEvent ev(type, &mev_data->ev);
                view->onMouseMoved(ev);
            }
            break;
        }
        case WIIU_MOUSE_UP:
        {
            WiiUMouseData *mev_data = static_cast<WiiUMouseData *>(data);

            view = (WebView *) mev_data->webview;
            if (view != NULL) {
                WebCore::PlatformMouseEvent ev(type, &mev_data->ev);
                view->onMouseReleased(ev);
            }
            break;
        }
        case WIIU_MOUSE_DOWN:
        {
            WiiUMouseData *mev_data = static_cast<WiiUMouseData *>(data);

            view = (WebView *) mev_data->webview;
            if (view != NULL) {
                WebCore::PlatformMouseEvent ev(type, &mev_data->ev);
                view->onMousePressed(ev);
            }
            break;
        }
        case WIIU_KEY_DOWN:
        {
            WiiUKeyboardData *kev_data = static_cast<WiiUKeyboardData*>(data);

            view = (WebView *) kev_data->webview;
            if (view != NULL) {
                WebCore::PlatformKeyboardEvent ev(type, &kev_data->ev);
                view->onKeyDown(ev);
            }
            break;
        }
        case WIIU_KEY_UP:
        {
            WiiUKeyboardData *kev_data = static_cast<WiiUKeyboardData*>(data);

            view = (WebView *) kev_data->webview;
            if (view != NULL) {
                WebCore::PlatformKeyboardEvent ev(type, &kev_data->ev);
                view->onKeyUp(ev);
            }
            break;
        }
        case WIIU_EXTENSION_EVENT:
        {
            //WebCore::JSIOFileExtension::DispatchIOEvent(data);
            JSEvent *pJsEvent = (JSEvent *)data;
            //JSSystemEventDispatcher::GetInstance()->DispatchEvent(*pJsEvent);

            JSSystemEventDispatcher::GetInstance()->DispatchSystemEvent(
                pJsEvent->ToString(),
                pJsEvent->ToString(),
                pJsEvent->GetData(),
                WIIU_EXTENSION_EVENT
            );
            break;
        }
        case WIIU_IO_EVENT:
        {
            JSEvent *pJsEvent = (JSEvent *)data;

            JSSystemEventDispatcher::GetInstance()->DispatchSystemEvent(
                pJsEvent->ToString(),
                JSSystemEventDispatcher::GetInstance()->GetEventString(
                    IO,
                    pJsEvent->ToString()
                ),
                pJsEvent->GetData(),
                WIIU_IO_EVENT
            );
            break;
        }
        case WIIU_AOC_EVENT:
        {
            JSEvent *pJsEvent = (JSEvent *)data;

            JSSystemEventDispatcher::GetInstance()->DispatchSystemEvent(
                pJsEvent->ToString(),
                JSSystemEventDispatcher::GetInstance()->GetEventString(
                    AOC,
                    pJsEvent->ToString()
                ),
                pJsEvent->GetData(),
                WIIU_AOC_EVENT
            );
            break;
        }
        case WIIU_ACCOUNT_SERVICE_TOKEN:
        {
            JSEvent *pJsEvent = (JSEvent *)data;

            JSSystemEventDispatcher::GetInstance()->DispatchSystemEvent(
                pJsEvent->ToString(),
                JSSystemEventDispatcher::GetInstance()->GetEventString(
                    ACT,
                    pJsEvent->ToString()
                ),
                pJsEvent->GetData(),
                WIIU_ACCOUNT_SERVICE_TOKEN
            );
            break;
        }
        case WIIU_SWKBD_EVENT:
        {
            if(swkbdEventCallback != NULL)
            {
                swkbdEventCallback(data);
            }
            //WebCore::JSSoftwareKeyboardExtension::DispatchSWKBDEvent(data);
            break;
        }
        case WIIU_USB_KEYBOARD_EVENT:
        {
        	JSSystemEventDispatcher::GetInstance()->QueueEventForDispatch(
				((WiiUUSBKeyboardData *)data)->Attached() ? "usbKeyboardConnected" : "usbKeyboardDisconnected",
				SYSTEM,
				CHANNEL_0,
				DEFAULT,
				0
            );
        	break;
        }
        case WIIU_RUN_ON_MAIN:
        {
            WTF::dispatchFunctionsFromMainThread();
            break;
        }
        case WIIU_SCROLL_RIGHT:
        {
            WiiUScrollEvent *ev = static_cast<WiiUScrollEvent*>(data);

            view = (WebView *) ev->webview;
            if (view != NULL) {
                view->scrollRight();
            }
            break;
        }
        case WIIU_SCROLL_UP:
        {
            WiiUScrollEvent *ev = static_cast<WiiUScrollEvent*>(data);

            view = (WebView *) ev->webview;
            if (view != NULL) {
                view->scrollUp();
            }
            break;
        }
        case WIIU_SCROLL_DOWN:
        {
            WiiUScrollEvent *ev = static_cast<WiiUScrollEvent*>(data);

            view = (WebView *) ev->webview;
            if (view != NULL) {
                view->scrollDown();
            }
            break;
        }
		case WIIU_LOAD_URL:
		{
			WiiULoadURLEvent *lurl_data = static_cast<WiiULoadURLEvent*>(data);
			view = (WebView *) lurl_data->webview;
			if (view != NULL)
			{
				view->load(lurl_data->url.c_str());
				view->setVisible(true);
			}	
			break;
		}
		case WIIU_SCREENSHOT:
		{
			WiiUScreenshotEvent *screenshot_data = static_cast<WiiUScreenshotEvent*>(data);
			if (screenshot_data->webview != NULL)
				WebViewSaveScreen(screenshot_data->webview, screenshot_data->filename.c_str());
			break;
		}
		case WIIU_SET_USER_AGENT:
		{
			WiiUSetUserAgentEvent *sua_data = static_cast<WiiUSetUserAgentEvent*>(data);
			view = (WebView *) sua_data->webview;
			if (view != NULL)
				view->SetUserAgent(sua_data->ua.c_str());
			break;
		}
		case WIIU_SET_CUSTOM_HEADER:
		{
			WiiUSetCustomHeaderEvent *sch_data = static_cast<WiiUSetCustomHeaderEvent*>(data);
			view = (WebView *) sch_data->webview;
			if (view != NULL)
				view->SetCustomHeader(sch_data->header.c_str());
			break;
		}
		case WIIU_SET_PROXY_INFO:
		{
			WiiUSetProxyInfoEvent *proxy_data = static_cast<WiiUSetProxyInfoEvent*>(data);
			
			WebView::setProxyInfo(proxy_data->host.c_str(), proxy_data->port, proxy_data->proxyType, proxy_data->username.c_str(), proxy_data->password.c_str());
			break;
		}
		case WIIU_SET_CERTIFICATE:
		{
			WiiUSetCertificateEvent *cert_data = static_cast<WiiUSetCertificateEvent*>(data);
			if (cert_data->bClientCert)
				WebView::setClientCertificate(cert_data->cert.c_str());
			else
				WebView::setCertificate(cert_data->cert.c_str());
			break;
		}
		case	WIIU_CREATE_VIEW:
		{
			WiiUCreateViewEvent *createView_data = static_cast<WiiUCreateViewEvent*>(data);
			OS::Report("WIIU_CREATE_VIEW %d x %d\n", createView_data->size->width, createView_data->size->height);

			*createView_data->webview = new WebView(createView_data->size->width, createView_data->size->height,
													createView_data->cbs, createView_data->paintCB, createView_data->settings, NULL, createView_data->bNoSetup,
													createView_data->userdata, createView_data->inspectorview, createView_data->bInactiveDueToExtDisplay);
			pthread_cond_broadcast(createView_data->cond);
			break;
		}
		case	WIIU_DESTROY_VIEW:
		{
			WiiUDestroyViewEvent *destroyView_data = static_cast<WiiUDestroyViewEvent*>(data);

			WebViewDestroy(destroyView_data->webview, true);
			pthread_cond_broadcast(destroyView_data->cond);
			break;
		}
		case WIIU_SET_TEXT_FIELD_EVENT:
/*
			WiiUSetTextFieldDataEvent *pEventData = static_cast<WiiUSetTextFieldDataEvent *>(data);
			WebCore::ExceptionCode exc = 0;
			WebCore::Node *pNode = (WebCore::Node *)pEventData->pNode;
			WebCore::CharacterData *pData = dynamic_cast<WebCore::CharacterData *>(pNode);
			if (pData)
				pData->setData(pEventData->data.c_str(), exc);
			else
				pNode->setNodeValue(pEventData->data.c_str(), exc);
*/
			break;
		case WIIU_BLUR_CURRENT:
		{
			WiiUBlurEvent * const event_data = static_cast<WiiUBlurEvent*>(data);
			view = static_cast<WebView*>(event_data->webview());
			view->blur();
			break;
		}
		case WIIU_ERROR_UPDATE:
		{
			JSEvent *pJsEvent = (JSEvent *)data;
			JSSystemEventDispatcher::GetInstance()->DispatchErrorEvent(pJsEvent);
			break;
		}
		case WIIU_FOCUS_LOST:
		{
			WiiUFocusEvent *focus_ev = static_cast<WiiUFocusEvent*>(data);

			view = (WebView *) focus_ev->webview;
			if (view != NULL) {
				view->focusLost();
			}

			break;
		}
		case WIIU_FOCUS_GAINED:
		{
			WiiUFocusEvent *focus_ev = static_cast<WiiUFocusEvent*>(data);

			view = (WebView *) focus_ev->webview;
			if (view != NULL) {
				view->focusGained();
			}

			break;
		}
		case WIIU_PREV_PAGE:
		{
			WiiUNavEvent *event_data = static_cast<WiiUNavEvent*>(data);

			view = (WebView *) event_data->webview;
			if (view != NULL) {
				view->goBack();
			}
			break;
		}
		case WIIU_NEXT_PAGE:
		{
			WiiUNavEvent *event_data = static_cast<WiiUNavEvent*>(data);

			view = (WebView *) event_data->webview;
			if (view != NULL) {
				view->goForward();
			}
			break;
		}
		case WIIU_REFRESH_PAGE:
		{
			WiiUNavEvent *event_data = static_cast<WiiUNavEvent*>(data);

			view = (WebView *) event_data->webview;
			if (view != NULL) {
				view->reload();
			}
			break;
		}
		case WIIU_STOP_LOAD:
		{
			WiiUNavEvent *event_data = static_cast<WiiUNavEvent*>(data);

			view = (WebView *) event_data->webview;
			if (view != NULL) {
				view->stop();
			}
			break;
		}
		case WIIU_LOAD_PAGE:
		{
			WiiUNavEvent *event_data = static_cast<WiiUNavEvent*>(data);

			view = (WebView *) event_data->webview;
			if (view != NULL) {
				view->load(event_data->url.c_str());
			}
			break;
		}
		case WIIU_CLEAR_PAGE:
		{
			WiiUNavEvent *event_data = static_cast<WiiUNavEvent*>(data);

			view = (WebView *) event_data->webview;
			if (view != NULL) {
				view->clear();
			}
			break;
		}
		case WIIU_PAUSE_MEDIA:
		{
			WiiUPauseMedia *event_data = static_cast<WiiUPauseMedia*>(data);

			WebView *view = (WebView*)WebViewTV();
			if (view) {
				if (event_data->m_shouldPause) view->pauseMedia();
				else view->resumeMedia();
			}
			for (unsigned i = 0; i < (unsigned)NumGamePadWebViews(); ++i) {
				view = (WebView*)WebViewGamePad(i);
				if (view) {
					if (event_data->m_shouldPause) view->pauseMedia();
					else view->resumeMedia();
				}
			}
			break;
		}
		case WIIU_SCROLL_LEFT:
		{
			WiiUScrollEvent *ev = static_cast<WiiUScrollEvent*>(data);

			view = (WebView *) ev->webview;
			if(view != NULL) {
				view->scrollLeft();
			}
			break;
		}

		#if ENABLE(INSPECTOR_SERVER)
		case WIIU_INSPECTOR_SERVER_ACCEPT:
		{
			WiiUWebInspectorServerAcceptEvent *ev = static_cast<WiiUWebInspectorServerAcceptEvent *>(data);
			#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
			if (ev->isHio) WebCore::TcpServerWiiU::HandleAcceptHIO(ev->connectionFd);
			else           WebCore::TcpServerWiiU::HandleAccept(ev->connectionFd);
			#else
			WebCore::TcpServerWiiU::HandleAccept(ev->connectionFd);
			#endif
			break;
		}
		case WIIU_INSPECTOR_SERVER_START:
		{
			WiiUWebInspectorServerStartEvent *ev = static_cast<WiiUWebInspectorServerStartEvent *> (data);
			String frontendUrl(ev->frontendUrl.c_str(), ev->frontendUrl.length());
			String hostUrl(ev->hostUrl.c_str(), ev->hostUrl.length());
			WebCore::WebInspectorServer::shared().setFrontendUrl(frontendUrl, hostUrl);
			WebCore::WebInspectorServer::shared().listen("127.0.0.1", ev->listenPort);
			break;
		}
		case WIIU_INSPECTOR_SERVER_STOP:
		{
			OS::Report("WebCore::WebInspectorServer::close\n"); 
			WebCore::WebInspectorServer::shared().close();
			break;
		}
		#endif
		case WIIU_INSPECTOR_ENABLE:
		{
			#if ENABLE(INSPECTOR)
			WiiUWebInspectorEnableEvent * ev = static_cast<WiiUWebInspectorEnableEvent*>(data);
			ev->inspectedView->enableWebInspector(ev->inspectorEnable);
			#endif
			break;
		}
		case WIIU_RESTART_WEBVIEWS:
		{
			WiiURestartWebviewsEvent * TheEvent = (WiiURestartWebviewsEvent *)data;

			if (WebKitIsBackground())
			{
				OS::Report("RESTART_WEBVIEWS event received, but ignored due to background state.\n");

				if (WebCore::g_pSendEnableReloadPacket)
					WebCore::g_pSendEnableReloadPacket();
				else
					OS::Report("RESTART_WEBVIEWS ERROR - Initialization failed.\n");

				break;
			}

			if (!WebKitRestartWebViews())
				WebKitManager::webViewManager().setRestartWebViews(true, 1000.0f, TheEvent->m_pConfigData, TheEvent->m_DataSize); 
			else
			{
				OS::Report("WARNING: Currenting restarting webviews, cannot restart while restarting.\n"); 
			}
			break;
		}
		case WIIU_BG_NOTIFICATION:
		case WIIU_FG_NOTIFICATION:
		case WIIU_SHUTDOWN_NOTIFICATION:
		{
			WiiUShutdownNotification *ev = static_cast<WiiUShutdownNotification*>(data);

			if (type == WIIU_SHUTDOWN_NOTIFICATION)
				WebKitResume();

			// Send notification out to all views...
			WebView *view = WebKitManager::webViewManager().GetTVWebView();
			if(view != NULL)
			{
				switch(type)
				{
					case	WIIU_BG_NOTIFICATION:		view->sendBGNotification(); break;
					case	WIIU_FG_NOTIFICATION:		view->sendFGNotification(); break;
					case	WIIU_SHUTDOWN_NOTIFICATION: view->sendShutdownNotification(); break;
				}
			}
			
			view = WebKitManager::webViewManager().GetInspectorWebView();
			if(view != NULL)
			{
				switch(type)
				{
					case	WIIU_BG_NOTIFICATION:		view->sendBGNotification(); break;
					case	WIIU_FG_NOTIFICATION:		view->sendFGNotification(); break;
					case	WIIU_SHUTDOWN_NOTIFICATION: view->sendShutdownNotification(); break;
				}
			}
		
			for(int i = 0; i < WebKitManager::webViewManager().GetNumGamePadWebViews(); i++)
			{
				view = WebKitManager::webViewManager().GetGamePadWebView(i);
				if (view)
				{
					switch(type)
					{
						case	WIIU_BG_NOTIFICATION:		view->sendBGNotification(); break;
						case	WIIU_FG_NOTIFICATION:		view->sendFGNotification(); break;
						case	WIIU_SHUTDOWN_NOTIFICATION: view->sendShutdownNotification(); break;
					}
				}
			}

			#ifdef _WIIU_
			// Continue shutdown...
			switch(type)
			{
				case	WIIU_BG_NOTIFICATION:		WebViewProcessSwitchNotifyDone(SHIM_SWITCH_TO_BG); break;
				case	WIIU_FG_NOTIFICATION:		WebViewProcessSwitchNotifyDone(SHIM_SWITCH_TO_FG); break;
				case	WIIU_SHUTDOWN_NOTIFICATION: WebViewProcessSwitchNotifyDone(SHIM_SWITCH_SHUTDOWN); break;
			}
			#endif
			break;
		}
		default:
		{
			OS::Report("RECEIVED UNKNOWN EVENT\n");
		} break;
	}

	// Done with the event, get rid of the data...
	delete data;
	return true;
}

// --------------------------------------------------------------------------------------------------
// -- WebKitManager ---------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------
/**
	Startup global systems for Webkit use with the given settings
*/
bool WebKitStartup(const WebKitCallbacks_t *pCallbacks, int prio, int core)
{
	ASSERT(!WebKitManager::s_pInstance);
	WebKitManager::s_pInstance = new WebKitManager();
	if (!WebKitManager::s_pInstance)
		return(false);
	#if !OS(WINDOWS)
	return(WebKitManager::s_pInstance->startup(pCallbacks, prio, core));
	#else
	MSG msg;
	PeekMessage(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE); // ensure the message queue is created
	DWORD recvThreadId = GetCurrentThreadId();
	return(WebKitManager::s_pInstance->startup(pCallbacks, prio, core, recvThreadId));
	#endif
}

// --------------------------------------------------------------------------------------------------
/**
	Shutdown global systems for Webkit use
*/
void WebKitShutdown(void)
{
	ASSERT(WebKitManager::s_pInstance);
	delete WebKitManager::s_pInstance;
	WebKitManager::s_pInstance = NULL;
}

void WebKitRemoveCustomRenderingCommand(WebKitCustomRenderData *dataP)
{
	if(WebKitManager::s_pInstance == NULL)
		return;
	WebKitManager::s_pInstance->removeCustomRenderingCommand(dataP);
}

void WebKitAddCustomRenderingCommand(WebKitCustomRenderData *dataP)
{
	if(WebKitManager::s_pInstance == NULL)
		return;
	WebKitManager::s_pInstance->addCustomRenderingCommand(dataP);
}

// --------------------------------------------------------------------------------------------------
/**
	Main thread rendering handler
*/
void WebKitRendering(void)
{
	if (WebKitManager::s_pInstance)
		WebKitManager::s_pInstance->rendering();
}

void WebKitRenderCustomEvents(void *colorBufferP, long width, long height, bool isGamePad)
{
	if(WebKitManager::s_pInstance)
		WebKitManager::s_pInstance->renderCustomEvents(colorBufferP, width, height, isGamePad);
}

void WebKitRenderCustomEventsHold(void)
{
	if(WebKitManager::s_pInstance)
		pthread_mutex_lock(&WebKitManager::s_pInstance->m_customRenderingMutex);
}

void WebKitRenderCustomEventsRelease(void)
{
	if(WebKitManager::s_pInstance)
		pthread_mutex_unlock(&WebKitManager::s_pInstance->m_customRenderingMutex);
}

// --------------------------------------------------------------------------------------------------
/**
	Check if the webkit thread has existed
*/
bool WebKitDone(void)
{
	if (!WebKitManager::s_pInstance)
		return(false);
	return(WebKitManager::s_pInstance->WebKitRunStopped());
}

// --------------------------------------------------------------------------------------------------
/**
	Check if webkit thread is busy shutting down
*/
bool WebKitShuttingDown(void)
{
	if (!WebKitManager::s_pInstance)
		return false;
	return WebKitManager::s_pInstance->m_bWebViewRunShuttingDown;
}

// --------------------------------------------------------------------------------------------------
/**
	Trigger shutdown process with webkit
*/
void WebKitTriggerShutdown(void)
{
	WebKitSendEvent(new WiiUWebInspectorServerStopEvent());
	WebKitSendEvent(new WiiUStopEventLoopEvent());
}

// --------------------------------------------------------------------------------------------------
/**
	Signal to webkit that main thread has stopped using webkit resources
*/
void WebKitSignalReadyForShutdown(void)
{
	WebKitManager::signalShutdown();
}

// --------------------------------------------------------------------------------------------------
/**
	Stall until webkit has finished shutting down
*/
void WebKitWaitForShutdown(void)
{
	WebKitManager::waitForShutdown();
}

// --------------------------------------------------------------------------------------------------
/**
*/
bool WebKitIsBackground(void)
{
	return WebKitManager::instance().GetIsBackground();
}

// --------------------------------------------------------------------------------------------------
/**
	Returns true if the webviews should be restarted
	@return A \a bool value that is true if the webviews should be restarted and false otherwise
*/
bool WebKitRestartWebViews()
{
	return WebKitManager::WebKitRestartViews();
}

float WebKitGetRestartWebViewsFadeOutTime(void)
{
	return WebKitManager::WebKitRestartViewsFadeOutTime();
}

// --------------------------------------------------------------------------------------------------
/**
	Sets the flag that determines whether or not the web views will be torn down and restarted. 
	@param A \a bool value that should be true if the web views should restarted or false if they shouldn't be. 
*/
void WebKitSetRestartWebViews(bool restart, float fadeTimeMs, const char * vConfigXML, const unsigned int vConfigXMLLength)
{
	WebKitManager::SetWebkitRestartViews(restart, fadeTimeMs, vConfigXML, vConfigXMLLength);
}

// --------------------------------------------------------------------------------------------------
/**
	Sets a flag that determines if a webview restart event can be handled
*/
void WebViewBlockRestartWebViews(bool block, int blockSource)
{
	WebKitManager::BlockWebkitRestartViews(block, blockSource);
}

// --------------------------------------------------------------------------------------------------
/**
	Startup webkit support systems
*/
#if OS(WINDOWS)
bool WebKitManager::startup(const WebKitCallbacks_t *pCallbacks, signed prio, signed core, DWORD recvThreadId)
#else
bool WebKitManager::startup(const WebKitCallbacks_t *pCallbacks, signed prio, signed core)
#endif
{
	ASSERT(pCallbacks);
	
	m_appPaint = pCallbacks->m_paintCB;
	m_acquireGpuFgResources = pCallbacks->m_acquireGpuFgResourcesCB;

	#if PLATFORM(WIIU)
	shim_SetRealloc(pCallbacks->m_reallocCB);
	#endif

	#if OS(WINDOWS)
	m_recvThreadId = recvThreadId;
	#endif

	m_bWebViewRunStopped = false;
	m_bInitialized = true;

	// Startup private graphics context while still in main thread
	// (needed because the graphics libs will not allow setup of the context shadow from a secondary core)
	GraphicsContextPlatformPrivate::startup(callDLCB);

	// Initialize message queue for rendering
	#if PLATFORM(WIIU)
	OS::InitMessageQueue(&m_renderQueue, m_renderQueueBuffer, RENDER_MSG_MAXNUM - 2);	// -2 to keep two entries free at all times (1 read, 1 write)
	#elif OS(WINDOWS)
	// No initialization necessary
	#else
	m_renderQueue = mq_open("renderQueue", O_CREAT | O_RDWR, 0700, NULL);
	#endif
	m_renderQueueEntryIndex = 0;


	// Startup the events
	wk_InitEvents();
	// "Get" an instance of the wiiu event manager to make sure it is created right here and now
	// (otherwise a race condition could lead to >1 instances being created!)
	WebCore::WiiUEventManager::instance();

	#if PLATFORM(WIIU) || OS(WINDOWS)
	// Use pthread to create the main webkit thread
	// (we use pthread rather then the WiiU OS directly because Webkit assumes it's running in a thread created by pthread!)
	pthread_attr_t attr;
	sched_param_t sparam;
	pthread_attr_init(&attr);
	sparam.sched_priority = prio;
	sparam.sched_affinity = core;
	pthread_attr_setschedparam(&attr, &sparam);
	pthread_attr_setstacksize(&attr, WEBKIT_STACK_SIZE);
	pthread_attr_setthreadname(&attr, "WebKit main thread");
	pthread_create(&m_workerThread, &attr, workerThread, (void *)this);
	pthread_attr_destroy(&attr);
	#else
	// FIXME?: ignores the given priority
	pthread_create(&m_workerThread, NULL, &workerThread, (void *)this);
	#endif

	// We're all started up!
	return true;
}

// --------------------------------------------------------------------------------------------------
/**
	Initialize webkit support services
*/
namespace WebCore {namespace HTMLNames {
	extern bool g_HTMLNamesInitialized;
	void ResetHTMLNamesInitGuard();}
}
namespace JSC{void ResetJSCThreadGuard();}
WebKitManager::WebKitManager()
{
	//We need to make sure that the main thread will be initialized
	//every time we have a new one of these.
// 	WTF::ResetMainThreadInitGuard();
// 	JSC::ResetJSCThreadGuard();
// 	WebCore::HTMLNames::ResetHTMLNamesInitGuard();
// 	WebCore::HTMLNames::g_HTMLNamesInitialized = false;

	m_bInitialized = false;
	m_bRunLoopPaused = false;
	m_bRunLoopQuit = false;
	m_bWebViewRunShuttingDown = false;
	pthread_mutex_init(&m_workerMutex, NULL);
	pthread_mutex_init(&m_workerShutdownMutex, NULL);
	pthread_cond_init(&m_workerEvent, NULL);
	pthread_cond_init(&m_workerShutdownEvent, NULL);

	pthread_mutex_init(&m_customRenderingMutex, NULL);
	m_customRenderingListP = NULL;

	m_bPanicExit = false;

	#ifdef _WIIU_
	if (shim_IsBackground())
	#else
	if (0)
	#endif
	{
		m_bBackground = true;
		m_bRedrawEnabled = false;
	}
	else
	{
		m_bBackground = false;
		m_bRedrawEnabled = true;
	}

	#if ENABLE(INSPECTOR_SERVER)
	WebCore::g_pAddWebInspectorServerReference = WebCore::AddWebInspectorServerReference;
	WebCore::g_pRemoveWebInspectorServerReference = WebCore::RemoveWebInspectorServerReference;
	#endif
	LibPlatform::OS::Report("[WebKitManager]InspectorServer Add/Rem 0x%X -- 0x%X\n", (unsigned int)WebCore::g_pAddWebInspectorServerReference, (unsigned int)WebCore::g_pRemoveWebInspectorServerReference);
}

// --------------------------------------------------------------------------------------------------
/**
	Shutdown webkit support services
*/
WebKitManager::~WebKitManager()
{
	if (m_bInitialized)
	{
		OS::Report("WebKitManager: shutting down!\n");

		// note: for now we have no explicit "kill trigger" here, but trust to only get here when the webkit thread shut down anyway
		void *ret;
		pthread_join(m_workerThread, &ret);

		OS::Report("Webkit worker thread has joined, continue teardown...\n");
		
		pthread_mutex_destroy(&m_workerMutex);
		pthread_mutex_destroy(&m_workerShutdownMutex);
		pthread_cond_destroy(&m_workerEvent);
		pthread_cond_destroy(&m_workerShutdownEvent);
		pthread_mutex_destroy(&m_customRenderingMutex);
		m_customRenderingListP = NULL;
		
		if (!m_bPanicExit)
			GraphicsContextPlatformPrivate::shutdown();
	}
}

// --------------------------------------------------------------------------------------------------
/**
	Interrupt WebViewRun and pause the worker thread
*/
void WebKitManager::pause()
{
	ASSERT(WebKitManager::s_pInstance);

	pthread_mutex_lock(&WebKitManager::s_pInstance->m_workerMutex);
	if (!WebKitManager::s_pInstance->m_bRunLoopQuit && !WebKitManager::s_pInstance->m_bRunLoopPaused)
	{
		pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
		WebKitSendEvent(new WiiUPauseMedia(true));
	}
	else pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);

	// make sure this event is sent before m_bRunLoopPaused is set
	// because WebKitSendEvent won't send the event if m_bRunLoopPaused is true
	WebKitSendEvent(new WiiUInterruptWebViewEvent());

	pthread_mutex_lock(&WebKitManager::s_pInstance->m_workerMutex);
	if (!WebKitManager::s_pInstance->m_bRunLoopQuit && !WebKitManager::s_pInstance->m_bRunLoopPaused)
		WebKitManager::s_pInstance->m_bRunLoopPaused = true;
	pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
}

// --------------------------------------------------------------------------------------------------
/**
	Resume from pause
*/
void WebKitManager::resume()
{
	ASSERT(WebKitManager::s_pInstance);
	pthread_mutex_lock(&WebKitManager::s_pInstance->m_workerMutex);
	if (WebKitManager::s_pInstance->m_bRunLoopPaused)
	{
		WebKitManager::s_pInstance->m_bRunLoopPaused = false;
		pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
		pthread_cond_broadcast(&WebKitManager::s_pInstance->m_workerEvent);
		WebKitSendEvent(new WiiUPauseMedia(false));
	}
	else
	{
		pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
	}

	#if !PLATFORM(WIIU)
	unsigned int ScreenIndex;

	//If we're unpausing, the last vsync events we sent were totally ignored.
	//We need to turn these back off so we can retry.
	for (ScreenIndex = 0; ScreenIndex < WIIUWINDOW_NUM_TYPES; ScreenIndex++)
		g_VSyncMsgSent[ScreenIndex] = false;
	
	#endif
}

#if !PLATFORM(WIIU)
// --------------------------------------------------------------------------------------------------
/**
	Ping the work thread
*/
void WebKitManager::workping()
{
	ASSERT(WebKitManager::s_pInstance);
	pthread_cond_broadcast(&WebKitManager::s_pInstance->m_workerEvent);
}
#endif

// --------------------------------------------------------------------------------------------------
/**
	Signal the work thread that the main thread is good to exit
*/
void WebKitManager::signalShutdown(void) {
	ASSERT(WebKitManager::s_pInstance);
	pthread_cond_signal(&WebKitManager::s_pInstance->m_workerShutdownEvent);
}

// --------------------------------------------------------------------------------------------------
/**
	Wait for the work thread to exit
*/
void WebKitManager::waitForShutdown(void) {
	ASSERT(WebKitManager::s_pInstance);
	pthread_mutex_lock(&WebKitManager::s_pInstance->m_workerMutex);
	if (!WebKitManager::s_pInstance->m_bRunLoopQuit)
		pthread_cond_wait(&WebKitManager::s_pInstance->m_workerEvent, &WebKitManager::s_pInstance->m_workerMutex);
	pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
}

// --------------------------------------------------------------------------------------------------
/**
	Interrupt WebViewRun (or resume, if work is paused) and stop the worker thread
*/
void WebKitManager::quit()
{
	ASSERT(WebKitManager::s_pInstance);
	pthread_mutex_lock(&WebKitManager::s_pInstance->m_workerMutex);
	if (!WebKitManager::s_pInstance->m_bRunLoopQuit)
	{
		WebKitManager::s_pInstance->m_bRunLoopQuit = true;
		if (WebKitManager::s_pInstance->m_bRunLoopPaused)
		{
			pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
			pthread_cond_broadcast(&WebKitManager::s_pInstance->m_workerEvent);
		}
		else
		{
			pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
			WebKitSendEvent(new WiiUInterruptWebViewEvent());
		}
	}
	else
	{
		pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
	}
}

// --------------------------------------------------------------------------------------------------
/**
	Enable or disable redraw for all known views

	@note This is for use with WiiU's BG mode. In BG mode we cannot allow any rendering. This will
		  issue a redraw message when switching back to "enabled" state to make up for any paints
		  that had been missed.
*/
void WebKitManager::SetEnableRedraw(bool bEnable)
{
	// Any change?
	if (m_bRedrawEnabled != bEnable)
	{
		// Yes. Setup & tell all views...
		m_bRedrawEnabled = bEnable;
		WebView *pV = m_webViewManager.GetTVWebView();
		if (pV)
		{
			pV->setRedrawEnable(bEnable);
			pV->setRootLayerNeedsDisplay();
			if (bEnable)
			{
				// Make sure we repaint the whole view - we may have suppressed valueable paints while in BG mode!
				PlatformIntRect rect;
				rect.x = pV->x();
				rect.y = pV->y();
				rect.width = pV->width();
				rect.height = pV->height();
				pV->setDirtyRect(rect);
				pV->setRedrawPending();
				WebCore::WiiUEventManager::instance()->putEvent(new WiiURedrawEvent((WebViewHandle_t)pV));
			}
		}
		for(int i=0; i<m_webViewManager.GetNumGamePadWebViews(); ++i)
		{
			pV = m_webViewManager.GetGamePadWebView(i);
			if (pV)
			{
				pV->setRedrawEnable(bEnable);
	            pV->setRootLayerNeedsDisplay();
				if (bEnable)
				{
					// Make sure we repaint the whole view - we may have suppressed valueable paints while in BG mode!
					PlatformIntRect rect;
					rect.x = pV->x();
					rect.y = pV->y();
					rect.width = pV->width();
					rect.height = pV->height();
					pV->setDirtyRect(rect);
					pV->setRedrawPending();
					WebCore::WiiUEventManager::instance()->putEvent(new WiiURedrawEvent((WebViewHandle_t)pV));
				}
			}
		}
	}
}

// --------------------------------------------------------------------------------------------------
/**
	Setup webkit state such, that other threads still can exit

	@note	This is only to be used with truely fatal error conditions like JIT OOMs etc.
*/
void WebKitManager::SetPanicExit(void)
{
	// Pretend we are in logical BG mode... We are not, but we don't execute anything anymore & we need to signal BG mode to
	// allow the main thread to exit. Hence... we "lie".
	m_bBackground = true;
	// ...and for all practical purposes: we are stopped!
	m_bWebViewRunStopped = true;
	// ...also claim we're done shutting down!
	m_bRunLoopQuit = true;
	pthread_mutex_unlock(&WebKitManager::s_pInstance->m_workerMutex);
	pthread_cond_broadcast(&WebKitManager::s_pInstance->m_workerEvent);

	// And finally: let all know that we are in "panic exit mode"!
	m_bPanicExit = true;
	m_webViewManager.SetPanicExit();
}

// --------------------------------------------------------------------------------------------------
/**
	Paint callback used by Webkit thread to push display request to main thread
*/
void WebKitManager::paintCB(const void *texture, WebViewHandle_t webview, void *user_data)
{
	ASSERT(s_pInstance);
	s_pInstance->paint(texture, webview, user_data);
}

// --------------------------------------------------------------------------------------------------
/**
	Paint callback used by Webkit thread to push display request to main thread
*/
void WebKitManager::paint(const void *pTexture, WebViewHandle_t webView, void *pUserData)
{
	ASSERT(m_bInitialized);

	RenderQueueEntry *pMsg = &m_renderQueueEntries[m_renderQueueEntryIndex++];
	if (m_renderQueueEntryIndex == RENDER_MSG_MAXNUM)
		m_renderQueueEntryIndex = 0;

	pMsg->m_type = RenderQueueEntry::TYPE_SHOW_TEXTURE;
	pMsg->m_showTexture.m_pTexture = pTexture;
	pMsg->m_showTexture.m_WebView = webView;
	pMsg->m_showTexture.m_pUserData = pUserData;

	#if PLATFORM(WIIU)
	OS::Message	msg;
	msg.message = pMsg;
	OS::SendMessage(&m_renderQueue, &msg, SYSTEM_MESSAGE_BLOCK);
	#elif OS(WINDOWS)
	#warning FIXME: potential to overflow the internal message buffer!
	#warning Need some message queue implementation that can block/wait when full!
	//printf("***WebKitManager::paint\n");
	//PostThreadMessage(m_recvThreadId, 0xWIIU, (WPARAM)0, (LPARAM)pMsg);
	m_appPaint(pTexture, webView, pUserData);
	#else
	mq_send(m_renderQueue, (const char *)pMsg, sizeof(RenderQueueEntry *), 0);
	#endif
}

// --------------------------------------------------------------------------------------------------
/**
	Call DL callback used by Webkit thread to push render request to main thread
*/
void WebKitManager::callDLCB(void *pAddr, unsigned size, unsigned long long *pTimeStamp, unsigned long long timeStampValue, bool bDrawSync)
{
	if (s_pInstance)
	{
		s_pInstance->callDL(pAddr, size, timeStampValue, pTimeStamp, bDrawSync);
	}
}

// --------------------------------------------------------------------------------------------------
/**
	Call DL callback used by Webkit thread to push render request to main thread
*/
void WebKitManager::callDL(void *pAddr, unsigned size, unsigned long long timeStampValue, unsigned long long *pTimeStamp, bool bDrawSync)
{
	ASSERT(m_bInitialized);

	RenderQueueEntry *pMsg = &m_renderQueueEntries[m_renderQueueEntryIndex++];
	if (m_renderQueueEntryIndex == RENDER_MSG_MAXNUM)
		m_renderQueueEntryIndex = 0;

	pMsg->m_type = RenderQueueEntry::TYPE_CALL_DL;
	pMsg->m_callDL.m_pAddr = pAddr;
	pMsg->m_callDL.m_size = size;
	pMsg->m_callDL.m_timeStampVal = timeStampValue;
	pMsg->m_callDL.m_pTimeStamp = pTimeStamp;
	pMsg->m_callDL.m_bDrawSync = bDrawSync;

	#if PLATFORM(WIIU)
	OS::Message	msg;
	msg.message = pMsg;
	OS::SendMessage(&m_renderQueue, &msg, SYSTEM_MESSAGE_BLOCK);
	#elif OS(WINDOWS)
	#warning FIXME: potential to overflow the internal message buffer!
	#warning Need some message queue implementation that can block/wait when full!
	PostThreadMessage(m_recvThreadId, 0xWIIU, (WPARAM)0, (LPARAM)pMsg);
	#else
	mq_send(m_renderQueue, (const char *)pMsg, sizeof(RenderQueueEntry *), 0);
	#endif
}

void WebKitManager::WebKitEmptyRenderQueue()
{
	if (!s_pInstance)
		return;

	//Steal all of the messages from the render queue and ignore them.
	#if PLATFORM(WIIU)
	OS::Message msg;
	while(OS::ReceiveMessage(&s_pInstance->m_renderQueue, &msg, SYSTEM_MESSAGE_NOBLOCK))
	{
		//Nothing to do here.
	}
	#endif
}

void WebKitManager::addCustomRenderingCommand(WebKitCustomRenderData *dataP)
{
	pthread_mutex_lock(&m_customRenderingMutex);
	// new head!
	if(m_customRenderingListP == NULL) {
		m_customRenderingListP = dataP;
		dataP->prevP = dataP->nextP = NULL;

	// Put it at the end of the list
	} else {
		for(WebKitCustomRenderData *i = m_customRenderingListP; i; i = i->nextP) {
			if(!i->nextP) {
				i->nextP = dataP;
				dataP->prevP = i;
				dataP->nextP = NULL;
			}
		}
	}
	pthread_mutex_unlock(&m_customRenderingMutex);
}

inline void WebKitManager::removeCustomRenderingCommandHelper(WebKitCustomRenderData *dataP) {
	// already empty
	if(m_customRenderingListP == NULL)
		return;

	// If data is at the head of the list
	if(dataP == m_customRenderingListP) {
		m_customRenderingListP = dataP->nextP;
		if(m_customRenderingListP)
			m_customRenderingListP->prevP = NULL;
	
	// If the data is at the end of the list
	} else if(dataP->nextP == NULL)	{
		dataP->prevP->nextP = NULL;

	// If the data is in the middle somewhere
	} else {
		dataP->prevP->nextP = dataP->nextP;
		dataP->nextP->prevP = dataP->prevP;
	}
}

void WebKitManager::removeCustomRenderingCommand(WebKitCustomRenderData *dataP)
{
	pthread_mutex_lock(&m_customRenderingMutex);
	this->removeCustomRenderingCommandHelper(dataP);
	pthread_mutex_unlock(&m_customRenderingMutex);
}

void WebKitManager::renderCustomEvents(void *colorBufferP, long width, long height, bool isGamePad)
{
	// custom rendering first!
	pthread_mutex_lock(&m_customRenderingMutex);
	for(WebKitCustomRenderData *curP = m_customRenderingListP; curP; curP = curP->nextP) {
		// If DoRender returns false then either the render failed
		// or it's only meant to be a single frame pass.
		// Either way remove it form the list.
		if(!curP->DoRender(colorBufferP, width, height, isGamePad)) {
			this->removeCustomRenderingCommandHelper(curP);
		}

		curP->DoneRender();
	}
	pthread_mutex_unlock(&m_customRenderingMutex);
}

// --------------------------------------------------------------------------------------------------
/**
	Main thread webkit render code
	
	@return	True if anything was processed, false otherwise

*/
void WebKitManager::rendering()
{
	ASSERT(m_bInitialized);

	#if OS(WINDOWS)
	static bool PrintedStatus = false;
	if (!PrintedStatus)
		printf("***WebKitManager::rendering() Thread ID: 0x%X\n", (unsigned int)GetCurrentThreadId());
	PrintedStatus = true;
	#endif
	
	// Process all render queue entries we currently have in order of submission...
	#if PLATFORM(WIIU)
	OS::Message	msg;
	while(OS::ReceiveMessage(&m_renderQueue, &msg, SYSTEM_MESSAGE_NOBLOCK))
	#elif OS(WINDOWS)
	MSG msg;
	while(PeekMessage(&msg, NULL, 0xWIIU, 0xWIIU, PM_REMOVE) != 0)
	#else
	char *msg = (char *)malloc(sizeof(RenderQueueEntry *));
	while(mq_receive(m_renderQueue, msg, sizeof(RenderQueueEntry *), 0) >= 0)
	#endif
	{
		#if PLATFORM(WIIU)
		RenderQueueEntry *pMsg = (RenderQueueEntry *)msg.message;
		#elif OS(WINDOWS)
		RenderQueueEntry *pMsg = (RenderQueueEntry *)msg.lParam;
		#else
		RenderQueueEntry *pMsg = *((RenderQueueEntry **)msg);
		#endif
		switch(pMsg->m_type)
		{
			case	RenderQueueEntry::TYPE_CALL_DL:
			{
				WKgfx::RenderWebkitDL(pMsg->m_callDL.m_pAddr, pMsg->m_callDL.m_size, pMsg->m_callDL.m_timeStampVal, pMsg->m_callDL.m_pTimeStamp, pMsg->m_callDL.m_bDrawSync);
				break;
			}
			case	RenderQueueEntry::TYPE_SHOW_TEXTURE:
			{
				// Make application show the contents of the texture for the specified view...
				m_appPaint(pMsg->m_showTexture.m_pTexture, pMsg->m_showTexture.m_WebView, pMsg->m_showTexture.m_pUserData);
				break;
			}
			default:
			{
				ASSERT(!"*** Unknown render queue entry type!");
			}
		}
	}
	#if !PLATFORM(WIIU) && !OS(WINDOWS)
	free(msg);
	#endif
}

// --------------------------------------------------------------------------------------------------
/**
	Webkit worker thread
*/
void* WebKitManager::workerThread(void *pArg)
{
	return(reinterpret_cast<WebKitManager*>(pArg)->_workerThread());
}


//////////////////////////////////////////////////////////////////////////
///	Name	: HandleVSync
///
///	Purpose	: If a vsync message has gone unhandled by the main event
///				loop, it will be processed here. (Simulator Only)
//////////////////////////////////////////////////////////////////////////
void WebKitManager::HandleVSync()
{
	//NOTE: This function is only relevant to the simulator build, since it technically
	//		runs faster than 60Hz. It's possible that a v-sync message may need to be
	//		handled outside of the main event loop.

	#if !PLATFORM(WIIU)
	if (g_IsInInspectorMode)
	{
		WebView *view = (WebView *)WebViewTV();
		if (view && g_VSyncMsgSent[WIIUWINDOW_INSPECTOR])
		{
			view->_ForcePaintCall();
			g_VSyncMsgMutex.lock();
			g_VSyncMsgSent[WIIUWINDOW_INSPECTOR] = 0;	//We've handled the vsync message, we can take another now
			g_VSyncMsgMutex.unlock();
			g_VSyncMsgCond.signal();
		}
	}
	else
	{
		for(int i = 0; i < WIIUWINDOW_NUM_TYPES; ++i)
		{
			// FIXME: this needs to be revised if/when the WiiUWindowType enum is updated! -TS-NSA 2012-06-14
			WebView *view = NULL;
			switch(i)
			{
				case WIIUWINDOW_MAIN: view = (WebView *)WebViewTV(); break;
				case WIIUWINDOW_GAMEPAD0: view = (WebView *)WebViewGamePad(0); break;
			};

			// We may need to manually trigger a redraw
			if(view && g_VSyncMsgSent[i])
			{
				view->_ForcePaintCall();
				g_VSyncMsgMutex.lock();
				g_VSyncMsgSent[i] = 0;	//We've handled the vsync message, we can take another now
				g_VSyncMsgMutex.unlock();
				g_VSyncMsgCond.signal();
			}
		}
	}
	#endif
}

// --------------------------------------------------------------------------------------------------
/**
	Handler for CRASH() macro calls under NDEBUG conditions
*/
#ifdef NDEBUG
static void WebKitFatalDoneCB(void)
{
	WebKitManager::instance().SetPanicExit();
	shim_LetMasterThreadDrop();
}

static void WebKitCrashCB(void)
{
	// If this is the webkit thread we must remove it from the process switching sync-registration right now!
	if (WebKitManager::isWebkitThread())
	{
		// No more syncing...
		shim_UnregisterSyncThread(NULL);

		// Trigger the display of the fatal error...
		LibPlatform::Error::GetInstance()->setFatalError(WebKitFatalDoneCB);
	}
	else
	{
		// Trigger the display of the fatal error...
		LibPlatform::Error::GetInstance()->setFatalError(NULL);
	}


	// We exit "dirty". This could be VERY dirty, hence we got to tell shim not to wait for anybody...
	shim_pthread_ShutdownFreely();

	// We can't do more then to just quit this thread now. Any further execution would be highly unstable...
	pthread_exit((void*)-1);
}
#endif

// --------------------------------------------------------------------------------------------------
/**
	Webkit worker thread
*/
extern void CreateShaderTextureMaps(void);	// from wkgfx_GC_CustomShaders.cpp
void* WebKitManager::_workerThread()
{
	#if PLATFORM(WIIU)
	CreateShaderTextureMaps();

	#ifdef NDEBUG
	// Setup webkit crash vector
	WebKitSetCrashFunction(WebKitCrashCB);
	#endif

	#if ENABLE(JIT)
	// Startup JIT memory manager explicitly (must be done here as all interactions are limit to core 1)
	WTF::wiiuJIT_StartupMemory();

	// Setup all we need to lock JIT access in BG mode
	JSC::JSGlobalData::initJITLocker(&WebViewHandleProcessSwitch);
	#endif
	#endif

	#if PLATFORM(WIIU)
	// Register this thread to be notified for process switches
	shim_RegisterSyncThread(NULL, NULL, NULL);

	// We may require FG resources...
	if (!shim_IsBackground())
		WebKitManager::instance().AcquireGpuFgResources(true);
	#endif


	bool bForceExit = false;

	// Process events until requested to exit...
	pthread_mutex_lock(&m_workerMutex);

	#if !PLATFORM(WIIU)
	//Initialize the renderer here, since all the graphics run on this thread
	WiiURenderer::ExternalInit();
	#endif

	while (!m_bRunLoopQuit && !bForceExit)
	{
		if (m_bRunLoopPaused)
		{
			if (SoftwareKeyboard::GetInstance()->IsDisplaying() || Error::GetInstance()->IsShowing())
			{
				struct timeval now;
				gettimeofday(&now, NULL);
				struct timespec until;
				until.tv_sec  = now.tv_sec;
				until.tv_nsec = now.tv_usec * 1000;
				until.tv_nsec += 15 * 1e6;	// add 15ms to the current time

				// if nanoseconds >= 1,000,000,000, carry the one
				if (until.tv_nsec >= 1e9)
				{
					until.tv_sec += 1;
					until.tv_nsec -= 1e9;
				}

				int e = pthread_cond_timedwait(&m_workerEvent, &m_workerMutex, &until);
				if(e == ETIMEDOUT)
				{
					pthread_mutex_unlock(&m_workerMutex);
					// it is still OK to process events in this case because if WebKit is paused,
					// only BG, FG, and SHUTDOWN notifications are sent and that is exactly what we
					// might need to process here - see WebKitSendEvent()
					WebKitProcessEvents();
					pthread_mutex_lock(&m_workerMutex);
				}
			}
			else
			{
				pthread_cond_wait(&m_workerEvent, &m_workerMutex);
			}
		}
		else
		{
			pthread_mutex_unlock(&m_workerMutex);

			//Let WebView do its thing.
			bForceExit = WebViewRun() ? true : false;

			pthread_mutex_lock(&m_workerMutex);
		}

		#if !PLATFORM(WIIU)
		//Handle any unhandled vsync messages
		if(!m_bRunLoopQuit && !bForceExit)
		{
			pthread_mutex_unlock(&m_workerMutex);
			HandleVSync();
			pthread_mutex_lock(&m_workerMutex);
		}
		#endif
	}
	pthread_mutex_unlock(&m_workerMutex);

	OS::Report("======== WebKit Thread Shutdown Starts =======\n");

	#if PLATFORM(WIIU)
	// Register this thread to be notified for process switches
	shim_UnregisterSyncThread(NULL);
	#endif

	//Signal that we're shutting down and wait until the main loop
	//tells us that it's done attempting to render stuff.
	m_bWebViewRunShuttingDown = true;
	pthread_mutex_lock(&m_workerShutdownMutex);
	pthread_cond_wait(&m_workerShutdownEvent, &m_workerShutdownMutex);

	//Destroy all of the WebViews and empty the event queue
	WebKitManager::webViewManager().DestroyWebViews(true);
	OS::Report("WebViews have been destroyed.\n");

	EmptyWebKitEventQueue();
	OS::Report("Event queue is done.\n");

	pthread_mutex_unlock(&m_workerShutdownMutex);
	
	#if PLATFORM(WIIU) && ENABLE(JIT)
	// Shut down JIT locker
	JSC::JSGlobalData* globalData = WebCore::JSDOMWindowBase::commonJSGlobalData();
	OS::Report("globalData->refCount(): %u \n", globalData->refCount());
	globalData->heap.destroy();
	JSC::JSGlobalData::exitJITLocker();
	//APW - 1/30/13: After old objects have been forcibly destroyed above, some may have queued shutdown functions.
	//This line was added to fix a bug with WebAudio cleanup specifically
	WTF::dispatchFunctionsFromMainThread();
	#endif

	#if PLATFORM(WIIU)
	// No more FG resources for us
	if (!shim_IsBackground())
		WebKitManager::instance().AcquireGpuFgResources(false);
	#endif

	//Mark the bits that need to be marked for proper exit
	#if !PLATFORM(WIIU)
	g_WaitForWebViewRun = false;
	#endif
	m_bWebViewRunStopped = true;

	//Tell the main loop that it is now safe to shutdown WebKit
	pthread_cond_signal(&m_workerEvent);

	// Exit thread...	
	return(NULL);
}

/**
	Sets the flag that determines whether or not the web views will be torn down and restarted. 
	@param A \a bool value that should be true if the web views should restarted or false if they shouldn't be. 
*/
void WebKitManager::SetWebkitRestartViews(bool bRestart, float fadeTimeMs, const char * vConfigXML, const unsigned int vConfigXMLLength)
{
	webViewManager().setRestartWebViews(bRestart, fadeTimeMs, vConfigXML, vConfigXMLLength);
}

/**
	Sets the blocked restart flag for the specified source that determines if a webview restart should be performed immediately
*/
void WebKitManager::BlockWebkitRestartViews(bool block, int blockSource)
{
	webViewManager().blockWebViewRestart(block, blockSource);
}

/**
	Returns true if the webviews should be restarted
	@return A \a bool value that is true if the webviews should be restarted and false otherwise
*/
bool WebKitManager::WebKitRestartViews() 
{
	return webViewManager().restartWebViews(); 
}

/**
	Returns the time in Milliseconds to fade the webkit to black
	@return A \a float value that is the time in milliseconds
*/
float WebKitManager::WebKitRestartViewsFadeOutTime(void)
{
	return webViewManager().getRestartWebViewFadeOutTime();
}

/**
	If we're currently waiting for webviews to restart, this will return the optional
	config.xml data to be used for updating our configuration settings during reload.
*/
const char * WebKitGetRestartWebViewsConfigXML()
{
	WebKitManager::webViewManager().getConfigXMLData();
}

/**
	Returns the length of the data returned from GetRestartWebViewsConfigXML()
*/
const unsigned int WebKitGetRestartWebViewsConfigXMLLength()
{
	WebKitManager::webViewManager().getConfigXMLDataLength();
}

void WebViewSetScreenshotCompleteCallback(void (*fn)(const char *, bool))
{
	WebViewManager::setScreenshotCompleteCallback(fn);
}

// --------------------------------------------------------------------------------------------------
#if !PLATFORM(WIIU)
#define SAVE_COMMON_DIR FS_SAVE_DIR"/common"
#define SAVE_ACCOUNT_DIR FS_SAVE_DIR"/account"
static void sanitizePath(std::string &path) {
	size_t pos = path.find('\\');
	while(pos != std::string::npos) {
		#if 0
		path.insert(pos, 1, '\\');
		pos = path.find('\\', pos+2);
		#else
		// The JS API appears to be absolutely intolerant of backward-slashes in native strings -TS-NSA 2012-04-24
		path.replace(pos, 1, 1, '/');
		pos = path.find('\\', pos);
		#endif
	}
}

void SetAssetsDir(const char *set) {
	if(set && *set) {
		std::string assetsdir = set;
		sanitizePath(assetsdir);
		WebCore::RenderThemeWiiU::setAssetsPath(assetsdir.c_str());
		WebCore::ScrollbarThemeWiiU::setAssetsPath(assetsdir.c_str());
		#if ENABLE(INSPECTOR)
		WebCore::InspectorClientWiiU::setAssetsPath(assetsdir.c_str());
		#endif
		#if ENABLE(INSPECTOR_SERVER)
		WebCore::WebInspectorServer::setAssetsPath(assetsdir.c_str());
		#endif
	}
	else {
		WebCore::RenderThemeWiiU::setAssetsPath(NULL);
		WebCore::ScrollbarThemeWiiU::setAssetsPath(NULL);
		#if ENABLE(INSPECTOR)
		WebCore::InspectorClientWiiU::setAssetsPath(NULL);
		#endif
		#if ENABLE(INSPECTOR_SERVER)
		WebCore::WebInspectorServer::setAssetsPath(NULL);
		#endif
	}
}

static std::string contentdir, savedir, savecommondir, savecommontempdir;

void SetContentDir(const char *set) {
	if(set && *set) {
		contentdir = set;
		//#if OS(WINDOWS)
		//mkdir(contentdir.c_str()); // try to create the directory if it doesn't exist
		//#else
		//mkdir(contentdir.c_str(), S_IRWXU); // try to create the directory if it doesn't exist
		//#endif

		sanitizePath(contentdir);

		FSSetMaskedDir(FS_DIR_VOLAPP, FS_CONTENT_DIR"/app/");
		FSSetSystemDir(FS_DIR_VOLAPP, contentdir.c_str());
		FSSetDirFlags(FS_DIR_VOLAPP, FS_DIR_FLAG_READ_WRITE, true);
	}
	else {
		contentdir.clear();
		FSSetMaskedDir(FS_DIR_VOLAPP, NULL);
		FSSetSystemDir(FS_DIR_VOLAPP, NULL);
		FSSetDirFlags(FS_DIR_VOLAPP, (unsigned long)FS_DIR_FLAG_ALL, false);
	}
}

void SetSaveDir(const char *set) {
	if(set && *set) {
		savedir = set;
		//#if OS(WINDOWS)
		//mkdir(savedir.c_str()); // try to create the directory if it doesn't exist
		//#else
		//mkdir(savedir.c_str(), S_IRWXU); // try to create the directory if it doesn't exist
		//#endif

		sanitizePath(savedir);

		FSSetMaskedDir(FS_DIR_APPACCOUNTSAVE, SAVE_ACCOUNT_DIR"/");
		FSSetSystemDir(FS_DIR_APPACCOUNTSAVE, savedir.c_str());
		FSSetDirFlags(FS_DIR_APPACCOUNTSAVE, FS_DIR_FLAG_READ_WRITE, true);
	}
	else {
		savedir.clear();
		FSSetMaskedDir(FS_DIR_APPACCOUNTSAVE, NULL);
		FSSetSystemDir(FS_DIR_APPACCOUNTSAVE, NULL);
		FSSetDirFlags(FS_DIR_APPACCOUNTSAVE, (unsigned long)FS_DIR_FLAG_ALL, false);
	}
}

void SetSaveCommonDir(const char *set) {
	if(set && *set) {
		savecommondir = set;
		//#if OS(WINDOWS)
		//mkdir(savecommondir.c_str()); // try to create the directory if it doesn't exist
		//#else
		//mkdir(savecommondir.c_str(), S_IRWXU); // try to create the directory if it doesn't exist
		//#endif

		sanitizePath(savecommondir);

		savecommontempdir = set;
		#if OS(WINDOWS)
		savecommontempdir.append("temp");
		mkdir(savecommontempdir.c_str()); // try to create the directory if it doesn't exist
		savecommontempdir.append("\\");
		#else
		savecommontempdir.append("temp");
		mkdir(savecommontempdir.c_str(), S_IRWXU); // try to create the directory if it doesn't exist
		savecommontempdir.append("/");
		#endif

		sanitizePath(savecommontempdir);

		FSSetMaskedDir(FS_DIR_APPCOMMONSAVE, SAVE_COMMON_DIR"/");
		FSSetSystemDir(FS_DIR_APPCOMMONSAVE, savecommondir.c_str());
		FSSetDirFlags(FS_DIR_APPCOMMONSAVE, FS_DIR_FLAG_READ_WRITE, true);

		FSSetMaskedDir(FS_DIR_APPTEMP, SAVE_COMMON_DIR"/temp/");
		FSSetSystemDir(FS_DIR_APPTEMP, savecommontempdir.c_str());
		FSSetDirFlags(FS_DIR_APPTEMP, FS_DIR_FLAG_READ_WRITE, true);
	}
	else {
		savecommondir.clear();
		savecommontempdir.clear();
		FSSetMaskedDir(FS_DIR_APPCOMMONSAVE, NULL);
		FSSetSystemDir(FS_DIR_APPCOMMONSAVE, NULL);
		FSSetDirFlags(FS_DIR_APPCOMMONSAVE, (unsigned long)FS_DIR_FLAG_ALL, false);
		FSSetMaskedDir(FS_DIR_APPTEMP, NULL);
		FSSetSystemDir(FS_DIR_APPTEMP, NULL);
		FSSetDirFlags(FS_DIR_APPTEMP, (unsigned long)FS_DIR_FLAG_ALL, false);
	}
}

const char *GetContentDir(void) {
	return contentdir.c_str();
}

const char *GetSaveDir(void) {
	return savedir.c_str();
}

const char *GetSaveCommonDir(void) {
	return savecommondir.c_str();
}

const char *GetSaveCommonTempDir(void) {
	return savecommontempdir.c_str();
}
#endif
