/*
 * Copyright (C) 2007 Apple Inc.  All rights reserved.
 * Copyright (C) 2008 INdT - Instituto Nokia de Tecnologia
 * Copyright (C) 2009-2010 ProFUSION embedded systems
 * Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef InspectorClientWiiU_h
#define InspectorClientWiiU_h

#include "InspectorClient.h"
#include "wk_appinterface.h"
#include "WebViewWiiU.h"
#include "InspectorFrontendClientLocal.h"
#include "Page.h"
#include <wtf/Forward.h>

#if !PLATFORM(WIIU)
#include <string>
#endif

namespace WebCore {
class Node;
class Page;

class InspectorClientWiiU : public WebCore::InspectorClient {
public:
    InspectorClientWiiU(WebView *webview, WebView *inspectorview);

    #if !PLATFORM(WIIU)
    // Set the base path for loading inspector assets
    static void setAssetsPath(const char *set);
    static std::string prependAssetsPath(const char *path);
    #endif

    virtual void inspectorDestroyed();

    virtual void openInspectorFrontend(InspectorController* controller);
	virtual void closeInspectorFrontend();
    virtual void bringFrontendToFront();
    virtual void highlight();
    virtual void hideHighlight();

    virtual void populateSetting(const String& key, String* value);
    virtual void storeSetting(const String& key, const String& value);

    virtual bool sendMessageToFrontend(const String&);

    Page * getInspectedPage();

#if ENABLE(INSPECTOR_SERVER)
    virtual void enableRemoteInspection();
    virtual bool hasRemoteFrontendConnected() const { return m_remoteFrontendConnected; }
    virtual void sendMessageToRemoteFrontend(const String& message);
    virtual void dispatchMessageFromRemoteFrontend(const String& message);
    virtual void remoteFrontendConnected();
    virtual void remoteFrontendDisconnected();
#endif

private:

    void initGraphics();
    WebView *m_inspectorView;
    WebView *m_webview;
#if ENABLE(INSPECTOR_SERVER)
    bool m_remoteFrontendConnected;
    int  m_remoteInspectionPageId;
#endif
};

class InspectorFrontendClientWiiU : public InspectorFrontendClientLocal {
public:
    InspectorFrontendClientWiiU( WebView* inspectedView, Page* inspectorPage);
    virtual ~InspectorFrontendClientWiiU();

    virtual void frontendLoaded();

    virtual String localizedStringsURL();

    virtual String hiddenPanels();

    virtual void bringToFront();
    virtual void closeWindow();


    virtual void attachWindow();
    virtual void detachWindow();

    virtual void setAttachedWindowHeight(unsigned height);

    virtual void inspectedURLChanged(const String& newURL);

    void inspectorClientDestroyed();

private:
    void updateWindowTitle();
    void destroyInspectorView(bool notifyInspectorController);

    WebView* m_inspectedView;
    Page*    m_inspectorPage;

    bool m_destroyingInspectorView;
    InspectorClientWiiU* m_inspectorClient;
#if ENABLE(INSPECTOR_SERVER)
    int m_remoteInspectionPageId;
#endif
};

}

#endif // InspectorClientWiiU_h
