
#ifndef PLATFORMSTATISTICS_H
#define PLATFORMSTATISTICS_H

#ifdef WIN32
	// Because Windows is so super-cool, it tries to define GetCurrentTime as GetTickCount.
	#ifdef GetCurrentTime
		#undef GetCurrentTime
	#endif
	#define GetCurrentTime GetCurrentTime
#endif

#include <map>
#include <string>

namespace LibPlatform {
	void ResetStatisticsGlobals();
	class PlatformStatistics {
		struct Timer {
			Timer(void) :
				started(false),
				start_time(0), 
				end_time(0), 
				avg_time(0),
				last_time(0) {}

			bool started;
			unsigned start_time;
			unsigned end_time;
			unsigned avg_time;
			unsigned last_time;
		};

		typedef bool Group;

	public:
		static PlatformStatistics *GetInstance(void);
		static void DestroyInstance(void);

		void StartTimer(std::string timer_name, std::string group_name);
		void StopTimer(std::string timer_name, std::string group_name);
		unsigned GetAverageTime(std::string timer_name);
		unsigned GetLastTime(std::string timer_name);
		unsigned GetCurrentTime(std::string timer_name);
		void EnableGroup(std::string switch_name);
		void DisableGroup(std::string switch_name);

	private:
		PlatformStatistics(void);
		~PlatformStatistics(void);

		std::map<std::string, Timer> timers;
		std::map<std::string, Group> groups;
	};
}

#endif//PLATFORMSTATISTICS_H
